from __future__ import annotations
from logging import LoggerAdapter
import threading
import json
from dataclasses import dataclass

import socketio

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from icefarm.client.lib import AbstractEventHandler

class EventLogger(LoggerAdapter):
    def process(self, msg, kwargs):
        return f"[EventServer] {msg}", kwargs

class SocketLogger(LoggerAdapter):
    def __init__(self, logger, url, extra = None):
        super().__init__(logger, extra)
        self.url = url

    def process(self, msg, kwargs):
        return f"[socket@{self.url}] {msg}", kwargs

@dataclass
class Event:
    serial: str
    event: str
    contents: dict

class EventServer:
    """
    Maintains websockets with the iCEFARM control and workers. Allows commands to be sent to the workers
    and dispatches received events to EventHandlers. EventHandlers receive events in the order they were added.
    """
    def __init__(self, client_id, eventhandlers: list[AbstractEventHandler], logger):
        self.client_id = client_id
        self.logger = EventLogger(logger)
        self.eventhandlers = eventhandlers

        self.worker_lock = threading.Lock()
        self.worker_sockets: dict[str, socketio.Client] = {}

        self.control_lock = threading.Lock()
        self.control_socket = None

    def addEventHandler(self, eh: AbstractEventHandler):
        """Adds an event handler. Should not be called after reservations have
        been made."""
        self.eventhandlers.append(eh)

    def handleEvent(self, event: Event):
        """
        Dispatches an Event to the eventhandlers. This may be called directly rather than as a result of a websocket message,
        but care needs to be taken not to produce recursive behavior.
        """
        for eh in self.eventhandlers:
            eh.handleEvent(event)

    def _createSocket(self, url):
        """Connects a socket to url and adds related handlers."""
        sio = socketio.Client()

        logger = SocketLogger(self.logger, url)

        @sio.event
        def connect():
            logger.info("connected")
        @sio.event
        def connect_error(_):
            logger.error("connection attempt failed")
        @sio.event
        def disconnect(reason):
            logger.error(f"disconnected: {reason}")
        @sio.event
        def event(data):
            try:
                msg = json.loads(data)
            except Exception:
                logger.error("received unparsable data")
                return

            serial = msg.get("serial")
            contents = msg.get("contents")

            if contents:
                event = contents.get("event")
            else:
                event = None

            if not serial or not event or not contents:
                logger.error("bad event contents")
                return

            logger.debug(f"received {event} event")
            event = Event(serial, event, contents)
            self.handleEvent(event)

        # TODO
        try:
            sio.connect(url, auth={"client_id": self.client_id}, wait_timeout=10)
            return sio
        except Exception:
            return False

    def connectWorker(self, url):
        if not self.control_socket:
            raise Exception("Control socket not connected")
        with self.worker_lock:
            if url in self.worker_sockets:
                return

            self.worker_sockets[url] = self._createSocket(url)

    def sendWorker(self, url, event, data: dict):
        """Sends data to worker socket."""
        try:
            data = json.dumps(data)
        except Exception:
            self.logger.error(f"failed to jsonify event {event} for worker {url}")
            return

        with self.worker_lock:
            sio = self.worker_sockets.get(url)

            if not sio:
                return False

            sio.emit(event, data)

            return True

    def connectControl(self, url):
        with self.control_lock:
            self.control_socket = self._createSocket(url)

    def disconnectWorker(self, url):
        with self.worker_lock:
            sio = self.worker_sockets.get(url)

            if not sio:
                return True

            sio.disconnect()

            del self.worker_sockets[url]

    def exit(self):
        for eh in self.eventhandlers:
            eh.exit()

        with self.worker_lock:
            urls = list(self.worker_sockets)

        for url in urls:
            self.disconnectWorker(url)



from __future__ import annotations
import threading
import uuid
import re
from dataclasses import dataclass
import time
import os

import serial

from icefarm.worker.device.state.core import AbstractState, FlashState, BrokenState
from icefarm.worker.device.state.reservable import reservable
from icefarm.utils.dev import get_devs

import typing
if typing.TYPE_CHECKING:
    from icefarm.worker.device import Device

#from  https://github.com/evolvablehardware/BitstreamEvolutionPico/blob/main/exampleProjectsC/bitstream_over_usb/bitstream_transfer_test.py
# TODO config file
BAUD = 115200            # ignored by TinyUSB but needed by pyserial
CHUNK_SIZE = 512         # bytes per write
INTER_CHUNK_DELAY = 0.00001  # seconds
BITSTREAM_SIZE = 0 #TODO

@dataclass
class Bitstream:
    location: str
    name: str
    batch_id: str

# TODO add flush amount
@reservable("pulsecount")
class PulseCountStateFlasher(AbstractState):
    def start(self):
        pulse_fac = lambda : PulseCountState(self.device)
        self.switch(lambda : FlashState(self.device, self.config.pulse_firmware_path, pulse_fac))

class PulseCountState(AbstractState):
    def __init__(self, state):
        super().__init__(state)

        self.cv = threading.Condition()
        self.bitstream_queue: list[Bitstream] = []
        # name -> pulses
        self.results = {}
        self.result_amount = 0
        # TODO configurable by client
        self.flush_threshold = 4

        # ensure new ports show correctly
        time.sleep(2)

        self.ser = self.connectSerial()
        self.reader = Reader(self.ser)
        self.sender = PulseCountEventSender(self.device_event_sender)

        self.exiting = False
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

        self.device_event_sender.sendDeviceInitialized()

    def connectSerial(self):
        paths = get_devs().get(self.serial)

        if not paths:
            self.switch(lambda : BrokenState(self.device))

        port = list(filter(lambda x : x.get("ID_USB_INTERFACE_NUM") == "00", paths))

        if not port:
            self.switch(lambda : BrokenState(self.device))

        port = port[0].get("DEVNAME")
        return serial.Serial(port, BAUD, timeout=0.1)

    @AbstractState.register("evaluate", "files", "batch_id")
    def queue(self, files, batch_id):
        media_path = self.device.media_path
        paths = [str(media_path.joinpath(str(uuid.uuid4()))) for _ in range(len(files))]

        for path, data in zip(paths, files.values()):
            with open(path, "wb") as f:
                f.write(data.encode("cp437"))
                f.flush()

        self.logger.debug(f"queued bitstreams: {list(files.keys())}")

        with self.cv:
            for path, name in zip(paths, files.keys()):
                self.bitstream_queue.append(Bitstream(path, name, batch_id))

            self.cv.notify_all()

        return True

    def run(self):
        time.sleep(2)

        while not self.exiting:
            with self.cv:
                if not self.bitstream_queue:
                    self.cv.wait_for(lambda : self.bitstream_queue or self.exiting)

                if self.exiting:
                    return

                bitstream = self.bitstream_queue.pop()

            self.logger.debug(f"evaluating bitstream {bitstream.name}")

            with open(bitstream.location, "rb") as f:
                data = f.read()

            data_len = len(data)

            self.reader.waitUntilReady()

            self.logger.debug(f"uploading bitstream {bitstream.name}")

            for i in range(0, data_len, CHUNK_SIZE):
                chunk = data[i:i+CHUNK_SIZE]
                self.ser.write(chunk)
                self.ser.flush()
                time.sleep(INTER_CHUNK_DELAY)

            self.logger.debug("waiting for pulse")

            result = self.reader.waitUntilPulse()

            self.logger.debug(f"got pulse: {result}")

            if result is False:
                with self.cv:
                    self.bitstream_queue.append(bitstream)
                    continue

            if bitstream.batch_id not in self.results:
                self.results[bitstream.batch_id] = []

            self.results[bitstream.batch_id].append((bitstream.name, result))
            self.result_amount += 1
            os.remove(bitstream.location)

            with self.cv:
                if not self.bitstream_queue or self.result_amount >= self.flush_threshold:
                    for batch_id, pulses in self.results.items():
                        if not self.sender.finished(batch_id, pulses):
                            self.logger.error("failed to send results")

                    self.results = {}
                    self.result_amount = 0

    def handleExit(self):
        self.exiting = True
        with self.cv:
            self.cv.notify_all()
        self.thread.join()
        self.reader.exit()
        self.ser.close()

    def reboot(self):
        self.switch(lambda : PulseCountStateFlasher(self.device))

class Reader:
    def __init__(self, port: serial.Serial):
        self.port = port
        self.cv = threading.Condition()
        self.ready = True
        self.last_pulse = None
        self.exiting = False

        self.thread = threading.Thread(target=self.read, daemon=True)
        self.thread.start()

    def read(self):
        last_read = ""
        while True:
            while self.port.is_open and not self.exiting and "\\r\\n" not in last_read:
                if (data := self.port.read(self.port.in_waiting or 1)):
                    last_read += str(data)[2:-1]

            if "\\r\\n" not in last_read:
                break

            line = last_read[:last_read.index("\\r\\n")]
            last_read = last_read[last_read.index("\\r\\n") + 4:]

            pulses = re.search("pulses: ([0-9]+)", line)
            if pulses:
                with self.cv:
                    self.last_pulse = pulses.group(1)
                    self.cv.notify_all()

            timeout = re.search("Watchdog timeout", line)
            if timeout:
                with self.cv:
                    self.last_pulse = False
                    self.cv.notify_all()

            wait = re.search("Waiting for bitstream transfer", line)
            if wait:
                with self.cv:
                    self.ready = True
                    self.cv.notify_all()

            if self.exiting:
                return

    def waitUntilReady(self):
        with self.cv:
            if not self.ready:
                self.cv.wait_for(lambda : self.ready)

            self.ready = False

    def waitUntilPulse(self):
        with self.cv:
            self.cv.wait_for(lambda : self.last_pulse is not None or self.exiting)
            last_pulse = self.last_pulse
            self.last_pulse = None
            return last_pulse

    def exit(self):
        self.exiting = True
        with self.cv:
            self.cv.notify_all()
        self.thread.join()

class PulseCountEventSender:
    def __init__(self, event_sender):
        self.event_sender = event_sender

    def finished(self, batch_id, pulses):
        return self.event_sender.sendDeviceEvent({
            "event": "results",
            "results": pulses,
            "batch_id": batch_id
        })



import threading
import sys
# TODO once control availability notifications are done,
# wait for those before returning. Currently a test exits immediately,
# so a redundant pico is needed to buffer default firmware flashing

def test_extend_all(client_fac):
    with client_fac() as client:
        assert client.extendAll() == client.getSerials()

def test_extend_serial(client_fac):
    with client_fac() as client:
        serials = client.getSerials()
        assert serials == client.extend(serials)

def test_end_all(client_fac):
    with client_fac() as client:
        client.endAll()
        assert not client.getSerials()

def test_end_serial(client_fac):
    with client_fac() as client:
        serials = client.getSerials()
        client.end(serials)
        assert not client.getSerials()

def test_evaluate_bitstreams(client_fac):
    with client_fac() as client:
        BITSTREAM_PATHS = ["examples/pulse_count_driver/precompiled_circuits/circuit_generated_2Khz.bin",
                    "examples/pulse_count_driver/precompiled_circuits/circuit_generated_8Khz.bin",
                    "examples/pulse_count_driver/precompiled_circuits/circuit_generated_32Khz.bin"]

        def timeout():
            sys.exit("Watchdog timeout")

        watchdog = threading.Timer(len(BITSTREAM_PATHS) * 20, timeout)
        watchdog.daemon = True
        watchdog.name = "watchdog-timeout"
        watchdog.start()

        pulses = list(client.evaluateBitstreams(BITSTREAM_PATHS))

        assert len(pulses) ==  3
        assert set(BITSTREAM_PATHS) == set(item[1] for item in pulses)



"""
Utility functions that don't fit the other modules.
"""
from __future__ import annotations
from logging import Logger
import re
import subprocess
import os
import inspect
import types
from configparser import ConfigParser
import itertools

from pexpect import fdpexpect

def get_env_default(var, default: str, logger: Logger):
    """Obtains an environment variable. If its not configured, it instead returns
    the default value and logs a warning message."""
    value = os.environ.get(var)

    if not value:
        value = default
        logger.warning(f"{var} not configured, defaulting to {default}")

    return value

def config_else_env(option: str, section: str, parser: ConfigParser, error=True, default=None):
    """Tries to find option from section of a .ini file. If it fails,
    it instead looks at the environment variable option. If this also
    fails, raises exception.."""
    if parser:
        if section in parser.sections():
            value = parser[section].get(option)
            if value:
                return value

    value = os.environ.get(option)
    if error:
        if not value:
            if default:
                return default
            raise Exception(f"Configuration error. Set {section}.{option} in the configuration or specify {option} as an environment variable.")

    return value

def check_default(devpath) -> bool:
    """Checks for whether a device is running the default firmware."""
    # Sometimes closing the fd takes a long time (> 10s) on some firmwares,
    # this might create issues. I'm not really sure what the cause is, I added
    # a read from stdio to the default firmware and it seems to fix the issue.
    # The same behavior happens from opening and closing the file in C.
    try:
        with open(devpath, "r") as f:
            p = fdpexpect.fdspawn(f, timeout=2)
            p.expect("default firmware", timeout=2)

    except Exception:
        return False

    return True

def get_ip() -> str:
    """Obtains local network ip from hostname -I."""
    res = subprocess.run(["hostname", "-I"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, check=True).stdout
    group = re.search("[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+", str(res))
    if group:
        return group.group(0)

def typecheck(fn, args) -> bool:
    """Checks whether args are valid types for fn. Only works on classes
    and non nested list generics. For dict, only checks if arg is a dict."""
    params = inspect.signature(fn).parameters.values()

    if len(params) != len(args):
        return False

    for arg, param in zip(args, params):
        annotation = param.annotation

        if annotation is inspect._empty:
            continue

        if inspect.isclass(annotation):
            if not isinstance(arg, annotation):
                return False

            continue

        if not isinstance(annotation, types.GenericAlias):
            return False

        if annotation.__origin__ is dict:
            continue

        if annotation.__origin__ is not list or not isinstance(arg, list):
            return False

        if len(annotation.__args__) != 1:
            return False

        type_ = annotation.__args__[0]

        for value in arg:
            if not isinstance(value, type_):
                return False

    return True

def json_to_args(json, parameters):
    values = list(map(json.get, parameters))
    if any(map(lambda x : x is None, parameters)):
        return False

    return values

def generate_circuit(hz, build_dir, build_script="src/icefarm/utils/build.sh", pcf_path="src/icefarm/utils/pico_ice.pcf", clk=48000000):
    """Builds a circuit of approximately hz."""
    if not os.path.isdir(build_dir):
        os.mkdir(build_dir)

    incr = int((clk/hz) - 1)
    veri = f"""
    module top (
        input CLK,
        output ICE_27,
        output LED_R,
        output LED_B
    );
    reg [22:0] counter;
    reg out;
    always @(posedge CLK) begin
        counter <= counter + 1;

        if (counter >= {incr}) begin
            out <= 1'b1;
            counter <= 23'b00000000000000000000000;
        end else begin
            out <= 1'b0;
        end
    end

    assign ICE_27 = out;
    assign LED_R = counter[22];
    endmodule
    """

    with open(os.path.join(build_dir, "top.v"), "w") as f:
        f.write(veri)

    subprocess.run(["bash", build_script, build_dir, pcf_path], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)

    return os.path.join(build_dir, "top.bin"), clk / incr / 1000

# Ideally would inherit dict but that doesn't behave nicely
class MappedQueues:
    """
    Manages a dict[Any, list[Any]] so that new lists are
    created if a key does not already exist. Empty list values
    mean that the corresponding key is treated as through it does
    not exist.
    """
    def __init__(self):
        self.state = {}

    def __getitem__(self, key) -> list:
        if key not in self.state:
            self.state[key] = []

        return self.state[key]

    def append(self, key, value):
        self[key].append(value)

    def pop(self, key, amount) -> list:
        """Pops up to amount items from queue key. If less
        than amount items are in queue key, all items will
        be popped.."""
        out = []

        for _ in range(amount):
            if key not in self:
                return out

            out.append(self[key].pop())

        return out

    def __bool__(self):
        return any(self.state.values())

    def __contains__(self, value):
        return bool(self.state.get(value))

    def __iter__(self):
        return itertools.takewhile(self.__getitem__, self.state)

    def keys(self):
        return list(iter(self))

    def values(self):
        return [self[key] for key in self]

def batch(l, amount):
    batches = [[] for _ in range(amount)]
    for i, item in enumerate(l):
        batches[i % amount].append(item)

    return batches



import json
import threading
from typing import List

import psycopg
from psycopg.types.enum import Enum, EnumInfo, register_enum

class DeviceStatus(Enum):
    available = 0
    reserved = 1
    await_flash_default = 2
    flashing_default = 3
    testing = 4
    broken = 5

class Database:
    """Base database class that syncs postgres enums with psycopg"""
    def __init__(self, dburl: str):
        self.url = dburl

        try:
            with psycopg.connect(self.url) as conn:
                info = EnumInfo.fetch(conn, "devicestatus")
                register_enum(info, conn, DeviceStatus)

        except Exception:
            raise Exception("Failed to connect to database")

    def execute(self, sql: str, args: tuple):
        try:
            with psycopg.connect(self.url) as conn:
                with conn.cursor() as cur:
                    cur.execute(sql, args)
                    if sql[0:4] != "CALL":
                        return cur.fetchall()

                    return True
        except Exception:
            return False

    def proc(self, sql: str, args: tuple):
        try:
            with psycopg.connect(self.url) as conn:
                with conn.cursor() as cur:
                    cur.execute(sql, args)
        except Exception:
            return False

        return True

    def getData(self, sql: str, args: tuple, columns: List[str], stringify=[]):
        if (data := self.execute(sql, args)) is False:
            return False

        out = list(map(lambda row : dict(zip(columns, row)), data))

        if stringify:
            for i, row in enumerate(out):
                for col in stringify:
                    out[i][col] = str(row[col])

        return out

    def listenReservations(self, callback):
        def l():
            with psycopg.connect(self.url, autocommit=True) as conn:
                conn.execute("LISTEN reservation_updates")
                gen = conn.notifies()

                for notif in gen:
                    try:
                        js = json.loads(notif.payload)
                        callback(js["device_id"], js["client_id"])
                    except Exception:
                        pass

        threading.Thread(target=l, daemon=True, name="reservation-update-listener").start()

    def listenAvailable(self, callback):
        def l():
            with psycopg.connect(self.url, autocommit=True) as conn:
                conn.execute("LISTEN device_available")
                gen = conn.notifies()

                for notif in gen:
                    try:
                        amount = notif.payload[1:-1]
                        callback(int(amount))
                    except Exception:
                        pass

        threading.Thread(target=l, daemon=True, name="devies-available-listener").start()



import threading
import time
import logging
from logging import Logger

import requests

class RemoteLogger:
    """Drop in replacement for a logging.Logger to also post to logs control server."""
    def __init__(self, logger: Logger, control_server: str, client_name: str, interval: int=30):
        self.logger: Logger = logger

        self.control_server = control_server
        self.client_name = client_name
        self.interval = interval

        self._backlog = []
        self._backlog_lock = threading.Lock()

        self._thread = threading.Thread(target=self._send, name="remote-logger", daemon=True)
        self._thread.start()

    def _send(self):
        while True:
            time.sleep(self.interval)
            with self._backlog_lock:
                logs = self._backlog
                self._backlog = []

            if not logs:
                continue

            try:
                res = requests.get(f"{self.control_server}/log", json={
                    "logs": logs,
                    "name": self.client_name
                }, timeout=10)
                if res.status_code != 200:
                    raise Exception
            except Exception:
                self.logger.error("[RemoteLogger] failed to send log results")

    def __getattr__(self, attr):
        # can't inherit from logging.Logger since its an externally managed singleton
        return getattr(self.logger, attr)

    def log(self, level, msg, *args, **kwargs):
        self.logger.log(level, msg, *args, **kwargs)
        with self._backlog_lock:
            self._backlog.append((level, msg))

    def debug(self, msg, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.log(logging.CRITICAL, msg, *args, **kwargs)



import time
import logging
import sys
import os

import pyudev
from flask import Flask
from flask_socketio import SocketIO

from icefarm.worker import Config, device

from icefarm.utils import RemoteLogger
from icefarm.worker.app import create_app, MAX_REQUEST_SIZE
from icefarm.worker.device import DeviceManager, DeviceEventSender
from icefarm.worker.device.state.core import FlashState, TestState, ReadyState
from icefarm.worker.device.state.reservable import PulseCountState

BITSTREAM_LENGTH = 104000

class FakeType(type):
    def __getattribute__(cls, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return FakeObject()

class FakeObject(metaclass=FakeType):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwds):
        return FakeObject()

    def __getattribute__(self, name):
        return FakeObject()

class FakeSerial:
    def __init__(self, *args, **kwargs):
        self.data_length = 0
        self.in_waiting = 0
        self.is_open = True

        self.queue = ""

    def write(self, data):
        self.data_length += len(data)
        if self.data_length >= BITSTREAM_LENGTH:
            self.queue += "'Bpulses: 12345\\r\\n'"
            self.queue += "'BWaiting for bitstream transfer\\r\\n'"

    def flush(self):
        pass

    def read(self, amount):
        if self.queue:
            msg = self.queue
            self.queue = ""
            return msg

        time.sleep(5)

    def close(self):
        pass

class FakeDevice:
    def __init__(self, properties):
        self.properties = properties

    def __iter__(self):
        return zip(self.properties.keys(), self.properties.values())

class FakeEventSender(DeviceEventSender):
    def __init__(self, event_sender, serial, logger):
        super().__init__(event_sender, serial, logger)
        self.events = []

    def sendDeviceEvent(self, contents):
        self.event_sender.append(contents)

def flash_state_start(self):
    """Replaces start during testing."""
    time.sleep(5)

    if self.timer:
        self.timer.cancel()
    self.switch(self.next_state_factory)

def test_state_start(self):
    if self.timer:
        self.timer.cancel()

    self.switch(lambda : ReadyState(self.device))

def patch(patch_event_sender=False):
    """Patches various modules to enable testing without hardware.
    - Disables pyudev events
    - Removes preexisting device scan from DeviceMonitor
    - FlashState succeeds after a few seconds and does not perform device interactions
    - TestState succeeds after a few seconds and does not perform device interactions
    - PulseCountState.ser: Serial is replaced with an emulation of the pulse count firmware
    """
    pyudev.Context = FakeObject
    pyudev.Monitor = FakeObject
    pyudev.MonitorObserver = FakeObject
    DeviceManager.scan = lambda self : None

    FlashState.start = flash_state_start
    FlashState.handleAdd = lambda self, dev : None
    TestState.start = test_state_start
    FlashState.handleAdd = lambda self, dev : None
    PulseCountState.connectSerial = lambda self : FakeSerial()

    if patch_event_sender:
        device.DeviceEventSender = FakeEventSender

def generate_device_add(serial):
    return FakeDevice({
        'CURRENT_TAGS': ':systemd:',
        'DEVLINKS': f'/dev/serial/by-path/pci-0000:00:14.0-usbv2-0:7:1.0 /dev/serial/by-path/pci-0000:00:14.0-usb-0:7:1.0 /dev/serial/by-id/usb-Raspberry_Pi_Pico_{serial}-if00',
        'DEVNAME': '/dev/ttyACM0',
        'DEVPATH': '/devices/pci0000:00/0000:00:14.0/usb1/1-7/1-7:1.0/tty/ttyACM0',
        'ID_BUS': 'usb',
        'ID_MM_CANDIDATE': '1',
        'ID_MODEL': 'Pico',
        'ID_MODEL_ENC': 'Pico',
        'ID_MODEL_ID': '0009',
        'ID_PATH': 'pci-0000:00:14.0-usb-0:7:1.0',
        'ID_PATH_TAG': 'pci-0000_00_14_0-usb-0_7_1_0',
        'ID_PATH_WITH_USB_REVISION': 'pci-0000:00:14.0-usbv2-0:7:1.0',
        'ID_REVISION': '0100',
        'ID_SERIAL': f'Raspberry_Pi_Pico_{serial}',
        'ID_SERIAL_SHORT': serial,
        'ID_TYPE': 'generic',
        'ID_USB_CLASS_FROM_DATABASE': 'Miscellaneous Device',
        'ID_USB_DRIVER': 'cdc_acm',
        'ID_USB_INTERFACES': ':020200:0a0000:',
        'ID_USB_INTERFACE_NUM': '00',
        'ID_USB_MODEL': 'Pico',
        'ID_USB_MODEL_ENC': 'Pico',
        'ID_USB_MODEL_ID': '0009',
        'ID_USB_PROTOCOL_FROM_DATABASE': 'Interface Association',
        'ID_USB_REVISION': '0100',
        'ID_USB_SERIAL': f'Raspberry_Pi_Pico_{serial}',
        'ID_USB_SERIAL_SHORT': serial,
        'ID_USB_TYPE': 'generic',
        'ID_USB_VENDOR': 'Raspberry_Pi',
        'ID_USB_VENDOR_ENC': 'Raspberry\\x20Pi',
        'ID_USB_VENDOR_ID': '2e8a',
        'ID_VENDOR': 'Raspberry_Pi',
        'ID_VENDOR_ENC': 'Raspberry\\x20Pi',
        'ID_VENDOR_ID': '2e8a',
        'MAJOR': '166',
        'MINOR': '0',
        'SUBSYSTEM': 'tty',
        'TAGS': ':systemd:',
        'USEC_INITIALIZED': '58315878783'
    })

def main():
    patch()
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))

    config_path = os.environ.get("USBIPICE_WORKER_CONFIG")
    if not config_path:
        config_path = None
    config = Config(path=config_path)

    if config.control_server_url:
        logger = RemoteLogger(logger, config.control_server_url, config.worker_name)

    app = Flask(__name__)
    socketio = SocketIO(app, max_http_buffer_size=MAX_REQUEST_SIZE)
    manager = create_app(app, socketio, config, logger)

    manager.handleDevEvent("add", generate_device_add("1111111111111111"))
    manager.handleDevEvent("add", generate_device_add("2222222222222222"))
    manager.handleDevEvent("add", generate_device_add("3333333333333333"))
    manager.handleDevEvent("add", generate_device_add("4444444444444444"))

    socketio.run(app, port=config.server_port, allow_unsafe_werkzeug=True, host="0.0.0.0")

if __name__ == "__main__":
    main()



from __future__ import annotations
from pathlib import Path
from logging import Logger, LoggerAdapter
import threading

from icefarm.worker.device import DeviceEventSender
from icefarm.worker.device.state.core import FlashState, TestState
from icefarm.worker.device.state.reservable import get_reservation_state_fac

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from icefarm.worker import WorkerDatabase, Config, EventSender
    from icefarm.worker.device import DeviceManager
    from icefarm.worker.device.state.core import AbstractState

# TODO add this to config
WORKER_MEDIA = "worker_media"

class DeviceLogger(LoggerAdapter):
    def __init__(self, logger, serial):
        super().__init__(logger, extra={"serial": serial})

    def process(self, msg, kwargs):
        return f"[{self.extra['serial']}] {msg}", kwargs

class Device:
    def __init__(self, serial: str, manager: DeviceManager, event_sender: EventSender, database: WorkerDatabase, logger: Logger):
        self.serial: str = serial
        self.manager: DeviceManager = manager
        self.database: WorkerDatabase = database
        self.logger: Logger = DeviceLogger(logger, self.serial)
        self.device_event_sender: DeviceEventSender = DeviceEventSender(event_sender, self.serial, self.logger)

        self._device: AbstractState = None
        self._device_lock = threading.RLock()

        self.path = Path(WORKER_MEDIA).joinpath(self.serial)

        self.path.joinpath("mount").mkdir(parents=True, exist_ok=True)
        self.path.joinpath("media").mkdir(exist_ok=True)

        self.__flashDefault()

    def __flashDefault(self):
        self.database.updateDeviceStatus(self.serial, "flashing_default")
        self.switch(lambda : FlashState(self, self.config.default_firmware_path, lambda : TestState(self), timeout=60))

    def handleDeviceEvent(self, action, dev):
        with self._device_lock:
            if not self._device:
                return

            with self._device.switching_lock:
                self.logger.debug(f'device {dev["DEVPATH"]}')

                if action == "add":
                    self._device.handleAdd(dev)
                    return

                if action == "remove":
                    self._device.handleRemove(dev)
                    return

                self.logger.warning(f"unhandled device action: {action}")

    def handleReserve(self, kind, args):
        fn = get_reservation_state_fac(self, kind, args)

        if not fn:
            return False

        self.switch(fn)
        return True

    def handleUnreserve(self):
        self.__flashDefault()
        return True

    def handleRequest(self, event, json):
        with self._device_lock:
            self._device.handleRequest(event, json)

    def handleExit(self):
        with self._device_lock:
            if self._device:
                self._device.handleExit()

    def reboot(self):
        self.logger.info("rebooting")
        with self._device_lock:
            if self._device:
                self._device.reboot()

    def switch(self, state_factory):
        with self._device_lock:
            if self._device:
                self._device.handleExit()
            device = state_factory()
            self._device = device
            self._device.start()

    @property
    def config(self) -> Config:
        return self.manager.config

    @property
    def mount_path(self) -> str:
        return self.path.joinpath("mount")

    @property
    def media_path(self) -> Path:
        return self.path.joinpath("media")



from icefarm.worker.device.state.core.AbstractState import AbstractState
from icefarm.worker.device.state.core.BrokenState import BrokenState
from icefarm.worker.device.state.core.FlashState import FlashState
from icefarm.worker.device.state.core.ReadyState import ReadyState
from icefarm.worker.device.state.core.TestState import TestState



from icefarm.control.ControlDatabase import ControlDatabase
from icefarm.control.ControlEventSender import ControlEventSender
from icefarm.control.Heartbeat import HeartbeatConfig, Heartbeat
from icefarm.control.Control import Control



from __future__ import annotations
from logging import Logger
import threading
from icefarm.client.lib import AbstractEventHandler, register

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from icefarm.client.lib import BaseAPI

class DefaultBaseEventHandler(AbstractEventHandler):
    @register("reservation ending soon", "serial")
    def handleReservationEndingSoon(self, serial: str):
        """Called when the reservation is almost finished."""

    @register("reservation end", "serial")
    def handleReservationEnd(self, serial: str):
        """Called when the reservation has ended."""

    @register("failure", "serial")
    def handleFailure(self, serial: str):
        """Called when the device experiences an unexpected failure
        that is not recoverable.
        """

class LoggerEventHandler(AbstractEventHandler):
    """Logs all events received by the EventServer."""
    def __init__(self, event_server, logger: Logger):
        super().__init__(event_server)
        self.logger = logger

    def handleEvent(self, event):
        self.logger.info(f"Received event: {event.event} serial: {event.serial} contents: {event.contents}")

class ReservationExtender(AbstractEventHandler):
    """Automatically extends device reservations when they are nearing completion."""
    def __init__(self, event_server, client, logger: Logger):
        super().__init__(event_server)
        self.client = client
        self.logger = logger

    @register("reservation ending soon", "serial")
    def handleReservationEndingSoon(self, serial: str):
        if self.client.extend([serial]):
            self.logger.info(f"refreshed reservation of {serial}")
        else:
            self.logger.error(f"failed to refresh reservation of device {serial}")

class AvailabilityWaiter(AbstractEventHandler):
    """Allows the client to sleep until a desired about of devices are available."""
    def __init__(self, event_server, client: BaseAPI):
        super().__init__(event_server)
        self.client = client
        self.last_available = 0
        self.cv = threading.Condition()

    @register("devices_available", "amount")
    def available(self, amount):
        self.last_available = amount
        with self.cv:
            self.cv.notify_all()

    def waitForAmountAvailable(self, amount):
        """
        Returns once at least amount devices are available for reservation. Note that
        this does not guarantee this client will be able to reserve them, as another client
        might first.
        """
        if self.client.available() >= amount:
            return

        with self.cv:
            self.cv.wait_for(lambda : amount <= self.last_available)



from __future__ import annotations
from logging import Logger
import threading

import requests

from icefarm.control import ControlDatabase
from icefarm.control.webapp import build_page

import typing
if typing.TYPE_CHECKING:
    from icefarm.control import ControlEventSender

class Control:
    def __init__(self, event_sender: ControlEventSender, database_url: str, logger: Logger):
        self.event_sender = event_sender
        self.database = ControlDatabase(database_url)
        self.logger = logger

        def update_available(amount):
            self.logger.debug(f"received notify for amount change: {amount}")
            self.event_sender.sendAll({
                "event": "devices_available",
                "amount": amount
            })

        self.database.listenAvailable(update_available)

        def reservation_end(serial, client):
            self.logger.info(f"received notify for reservation end device {serial} client {client}")
            self.event_sender.sendDeviceReservationEnd(serial, client)

        self.database.listenReservations(reservation_end)

    # TODO this feels out of place
    def getApp(self):
        return build_page(self.database)

    def extend(self, client_id: str, serials: list[str]) -> list[str]:
        return self.database.extend(client_id, serials)

    def extendAll(self, client_id: str) -> list[str]:
        return self.database.extendAll(client_id)

    def reboot(self, serials: list[str]):
        out = []
        for serial in serials:
            if not (url := self.database.getDeviceWorkerUrl(serial)):
                return False

            try:
                res = requests.get(f"{url}/reboot", json={
                    "serial": serial
                }, timeout=10)

                if res.status_code != 200:
                    raise Exception

                out.append(serial)

            except Exception:
                self.logger.warning(f"[Control] failed to send reboot command to worker {url} device {serial}")

        return out

    def delete(self, serials: list[str]):
        out = []
        for serial in serials:
            if not (url := self.database.getDeviceWorkerUrl(serial)):
                return False

            try:
                res = requests.get(f"{url}/delete", json={
                    "serial": serial
                    }, timeout=10)

                if res.status_code != 200:
                    raise Exception

                out.append(serial)
            except Exception:
                self.logger.warning(f"[Control] failed to send delete command to worker {url} device {serial}")

        return out

    def end(self, client_id: str, serials: list[str]) -> list[str]:
        data = self.database.end(client_id, serials)
        return list(map(lambda row : row["serial"], data))


    def endAll(self, client_id: str) -> list[str]:
        data = self.database.endAll(client_id)
        return list(map(lambda row : row["serial"], data))

    def getAmountAvailable(self):
        if (amount := self.database.getAmountAvailable()) is False:
            return False

        return {
            "amount": amount
        }

    def getDevicesAvailable(self):
        return self.database.getDevicesAvailable()

    def _sendReservationNotifications(self, con_info, kind, args):
        for row in con_info:
            def send_reserve():
                ip = row["ip"]
                port = row["serverport"]
                serial = row["serial"]

                try:
                    res = requests.get(f"http://{ip}:{port}/reserve", json={
                        "serial": serial,
                        "kind": kind,
                        "args": args
                    }, timeout=15)

                    if res.status_code != 200:
                        raise Exception
                except Exception:
                    pass

            thread = threading.Thread(target=send_reserve, name="send-reservation")
            thread.start()

    def reserve(self, client_id: str, amount: int, kind: str, args: dict) -> dict:
        if (con_info := self.database.reserve(amount, client_id, kind)) is False:
            return False

        self._sendReservationNotifications(con_info, kind, args)
        return con_info

    def reserveSerials(self, client_id: str, serials: list[str], kind: str, args: dict) -> dict:
        if (con_info := self.database.reserveSerials(client_id, serials, kind)) is False:
            return False

        self._sendReservationNotifications(con_info, kind, args)
        return con_info



from __future__ import annotations
from threading import Lock
from logging import Logger

import requests

class ConnectionInfo:
    """Database for information required to establish and maintain worker connections."""
    def __init__(self, ip: str, server_port: str):
        self.ip = ip
        self.server_port = server_port

    def __eq__(self, value):
        return self.ip == value.ip and self.server_port == value.ip

    def url(self):
        return f"http://{self.ip}:{self.server_port}"

class BaseAPI:
    """Provides an an abstraction over control server http endpoints and tracks reserved devices."""
    def __init__(self, url: str, client_name: str, logger: Logger):
        self.url = url
        self.name = client_name
        self.connection_info = {}
        self.lock = Lock()
        self.logger = logger

    def addSerial(self, serial, conn_info: ConnectionInfo):
        with self.lock:
            self.connection_info[serial] = conn_info

    def removeSerial(self, serial: str) -> bool:
        """Manually removes a device. Should be called after reservations end or
        devices fail. Returns whether it was successful."""
        with self.lock:
            if serial in self.connection_info:
                del self.connection_info[serial]
                return True

        return False

    def getSerials(self) -> list[str]:
        """Returns tracked serials."""
        with self.lock:
            return list(self.connection_info.keys())

    def getConnectionInfo(self, serial: str) -> ConnectionInfo:
        """Returns connection information associated with a serial."""
        with self.lock:
            return self.connection_info.get(serial)

    def usingConnection(self, info: ConnectionInfo):
        """Returns whether a connection is currently in use by a device."""
        with self.lock:
            return info in self.connection_info.values()

    def request(self, url: str, endpoint: str, json: dict, files=None) -> dict:
        """Sends a GET to url/endpoint with json. If files is specified, the data
        is instead sent as a multipart forum."""
        try:
            if files:
                res = requests.get(f"{url}/{endpoint}", data=json, files=files, timeout=20)
            else:
                res = requests.get(f"{url}/{endpoint}", json=json, timeout=20)

            if res.status_code != 200:
                self.logger.error(f"failed to GET /{endpoint}")
                return False

            return res.json()
        except Exception as e:
            self.logger.error(f"Exception during GET /{endpoint}: {e}")
            return False

    def requestControl(self, endpoint: str, json: dict) -> dict:
        """Sends a GET request to the specified endpoint of the control server with the specified json. Returns json of the response.
        Returns False on error."""
        return self.request(self.url, endpoint, json)

    def _addConnectionData(self, data):
        if data is False:
            return False

        out = []

        for row in data:
            serial = row["serial"]
            info = ConnectionInfo(row["ip"], row["serverport"])
            self.addSerial(serial, info)

            out.append(serial)

        return out

    def reserve(self, amount: int, kind: str, args: dict) -> dict:
        """Reserves amount devices with subscription_url as a event server.
        Returns successful reservations as a dict of serial -> bus"""
        json = {
            "amount": amount,
            "name": self.name,
            "kind": kind,
            "args": args
        }

        return self._addConnectionData(self.requestControl("reserve", json))

    def reserveSpecific(self, serials: list[str], kind: str, args: dict) -> dict:
        json = {
            "name": self.name,
            "kind": kind,
            "args": args,
            "serials": serials
        }

        return self._addConnectionData(self.requestControl("reserveserials", json))

    def available(self) -> int:
        data = self.requestControl("available", {})
        if not data:
            return False

        return data["amount"]

    def availableDevices(self) -> list[str]:
        return self.requestControl("devices", {})

    def extend(self, serials: list[str]) -> list[str]:
        """Extends the reservation on serials for by hour. The serials must be reserved under the client's name.
        Returns the serials that were extended."""
        return self.requestControl("extend", {
            "name": self.name,
            "serials": serials
        })

    def extendAll(self) -> list[str]:
        """Extends all reservations by the under the client's name for by hour. Returns the serials that were extended."""
        return self.requestControl("extendall", {
            "name": self.name
        })

    def end(self, serials: list[str]) -> list[str]:
        """Ends the reservation on serials. The serials must be reserved under the client's name. Returns the serials of reservations
        that were ended."""
        if not isinstance(serials, list):
            serials = list(serials)

        json = self.requestControl("end", {
            "name": self.name,
            "serials": serials
        })

        if json is False:
            return False

        for serial in json:
            self.removeSerial(serial)

        return json

    def endAll(self) -> list[str]:
        """Ends all reservations under the client's name. Returns the serials of reservations that were ended.."""
        json = self.requestControl("endall", {
            "name": self.name
        })

        if json is False:
            return False

        for serial in json:
            self.removeSerial(serial)

        return json



"""Starts the worker."""
import os
import logging
import sys
import threading
import json

from flask import Flask, Response
from flask_socketio import SocketIO
from socketio import ASGIApp
from asgiref.wsgi import WsgiToAsgi

from icefarm.worker import Config, WorkerDatabase
from icefarm.worker.device import DeviceManager

from icefarm.utils import EventSender
from icefarm.utils import RemoteLogger
from icefarm.utils.web import SyncAsyncServer, flask_socketio_adapter_connect, flask_socketio_adapter_on, inject_and_return_json

# 100 bitstreams
MAX_REQUEST_SIZE = 104.2 * 8000 * 100

def create_app(app: Flask, socketio: SocketIO | SyncAsyncServer, config: Config, logger: logging.Logger):

    event_sender = EventSender(socketio, config.libpg_string, logger)
    database = WorkerDatabase(config, logger)
    manager = DeviceManager(event_sender, database, config, logger)

    def handle_res_end(serial, client_id):
        if manager.unreserve(serial):
            database.handleReservationChange()

    database.listenReservations(handle_res_end)

    sock_id_to_client_id = {}
    id_lock = threading.Lock()

    @app.get("/heartbeat")
    def heartbeat():
        return Response(status=200)

    @app.get("/reserve")
    @inject_and_return_json
    def reserve(serial: str, kind: str, args: dict):
        return manager.reserve(serial, kind, args)

    @app.get("/reboot")
    @inject_and_return_json
    def reboot(serial: str):
        return manager.reboot(serial)

    @app.get("/delete")
    @inject_and_return_json
    def delete(serial: str):
        return manager.delete(serial)

    @socketio.on("connect")
    @flask_socketio_adapter_connect
    def connection(sid, environ, auth):
        client_id = auth.get("client_id")
        if not client_id:
            logger.warning("socket connection without client id")
            return

        with id_lock:
            sock_id_to_client_id[sid] = client_id

        event_sender.addSocket(sid, client_id)

    @socketio.on("disconnect")
    @flask_socketio_adapter_on
    def disconnect(sid, reason):
        with id_lock:
            client_id = sock_id_to_client_id.pop(sid, None)

        if not client_id:
            logger.warning("disconnected socket had no known client id")
            return

        event_sender.removeSocket(client_id)

    @socketio.on("request")
    @flask_socketio_adapter_on
    def handle(sid, data):
        with id_lock:
            client_id = sock_id_to_client_id.get(sid)

        if not client_id:
            logger.warning("socket sent request but has no known client id")
            return

        try:
            data = json.loads(data)
        except Exception:
            logger.error(f"failed to load json string from client {client_id}")
            return

        serial = data.get("serial")
        event = data.get("event")
        contents = data.get("contents")

        if not serial or not event or not contents:
            logger.error(f"bad request packet from client {client_id}")
            return

        if isinstance(serial, list):
            for s in serial:
                thread = threading.Thread(target=lambda : manager.handleRequest(s, event, contents), name="socket-request")
                thread.start()

        else:
            manager.handleRequest(serial, event, contents)

    @socketio.on("graceful_shutdown")
    @flask_socketio_adapter_on
    def shutdown(sid, data):
        database.enableShutDown()

        # this seems really dumb but the SyncAsyncServer adapter uses creates a new event
        # loop, so calling this directly results in nested event loops :/
        # TODO we should really adapt to the flask_socketio interface instead of socketio-server
        logger.warning("Graceful shutdown initilized")
        threading.Thread(target=lambda : socketio.emit("graceful_shutdown_initilized"), name='graceful-shutdown-notifier', daemon=True).start()

        def monitor():
            database.waitUntilNoReservations()
            manager.onExit()
            logger.info("Graceful shutdown complete")
            socketio.emit("shutdown_complete", "")

        threading.Thread(target=monitor, name="shutdown-monitor", daemon=True).start()

    return manager

def run_debug():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))

    config_path = os.environ.get("USBIPICE_WORKER_CONFIG")
    if not config_path:
        config_path = None
    config = Config(path=config_path)

    if config.control_server_url:
        logger = RemoteLogger(logger, config.control_server_url, config.worker_name)

    app = Flask(__name__)
    socketio = SocketIO(app, max_http_buffer_size=MAX_REQUEST_SIZE)
    create_app(app, socketio, config, logger)
    socketio.run(app, port=config.server_port, allow_unsafe_werkzeug=True, host="0.0.0.0")

def run_uvicorn():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    log_loc = os.environ.get("USBIPICE_WORKER_LOGS")
    if log_loc:
        logger.addHandler(logging.FileHandler(log_loc))
    else:
        raise Exception("USBIPICE_WORKER_LOGS not configured")

    logger.addHandler(logging.StreamHandler(sys.stdout))

    config_path = os.environ.get("USBIPICE_WORKER_CONFIG")
    if not config_path:
        config_path = None
    config = Config(path=config_path)

    if config.control_server_url:
        logger = RemoteLogger(logger, config.control_server_url, config.worker_name)

    app = Flask(__name__)
    socketio = SyncAsyncServer(async_mode="asgi", max_http_buffer_size=MAX_REQUEST_SIZE)
    create_app(app, socketio, config, logger)
    app = WsgiToAsgi(app)

    return ASGIApp(socketio, app)

if __name__ == "__main__":
    run_debug()



from __future__ import annotations
from logging import LoggerAdapter
from importlib.metadata import version
import threading

from icefarm.utils import Database
from icefarm.worker.device.state.reservable import get_registered_reservables

import typing
if typing.TYPE_CHECKING:
    from icefarm.worker import Config
    from icefarm.utils import DeviceStatus

class WorkerDataBaseLogger(LoggerAdapter):
    def process(self, msg, kwargs):
        return f"[WorkerDatabase] {msg}", kwargs

class WorkerDatabase(Database):
    # TODO use Database.exec
    """Provides access to database operations related to the worker process."""
    def __init__(self, config: Config, logger):
        super().__init__(config.libpg_string)
        self.worker_name = config.worker_name
        self.logger = WorkerDataBaseLogger(logger)
        self.cv = threading.Condition()

        usbipice_version = version("icefarm")
        reservables = get_registered_reservables()

        args = (self.worker_name, config.virtual_ip, config.virtual_server_port, usbipice_version, reservables)
        if not self.execute("CALL add_worker(%s::varchar(255), %s::varchar(255), %s::int, %s::varchar(255), %s::varchar(255)[])", args):
            raise Exception(f"Failed to add worker {self.worker_name}")

    def addDevice(self, deviceserial: str) -> bool:
        """Add a device to the database."""
        if not self.execute("CALL add_device(%s::varchar(255), %s::varchar(255))", (deviceserial, self.worker_name)):
            self.logger.error(f"failed to add device {deviceserial}")
            return False

        return True

    def updateDeviceStatus(self, deviceserial: str, status: DeviceStatus) -> bool:
        """Updates the status field of a device."""
        if not self.execute("CALL update_device_status(%s::varchar(255), %s::devicestatus)", (deviceserial, status)):
            self.logger.error(f"failed to update device {deviceserial} to status {status}")
            return False

        return True

    def enableShutDown(self):
        if not self.execute("CALL shutdown_worker(%s::varchar(255))", (self.worker_name,)):
            self.logger.error("Failed to enable shut down mode")
            return False

        return True

    def hasReservations(self):
        if not (data := self.execute("SELECT * FROM has_reservations(%s::varchar(255))", (self.worker_name,))):
            self.logger.error("Failed to check for reservations")
            return None

        return data[0][0]

    def handleReservationChange(self):
        with self.cv:
            self.cv.notify_all()

    def waitUntilNoReservations(self):
        if self.hasReservations():
            with self.cv:
                self.cv.wait_for(lambda : not self.hasReservations())

    def onExit(self):
        """Removes the worker and all related devices from the database."""
        if not self.execute("SELECT * FROM remove_worker(%s::varchar(255))", (self.worker_name,)):
            self.logger.warning(f"failed to remove worker {self.worker_name} before exit")



from icefarm.worker.device.state.core import AbstractState

class BrokenState(AbstractState):
    def __init__(self, state):
        super().__init__(state)
        self.database.updateDeviceStatus(self.serial, "broken")
        self.logger.error("device is broken")
        self.device_event_sender.sendDeviceFailure()



"""Abstracts classes for creating device drivers."""
from icefarm.client.lib.AbstractEventHandler import AbstractEventHandler, register
from icefarm.client.lib.EventServer import EventServer, Event
from icefarm.client.lib.BaseAPI import BaseAPI
from icefarm.client.lib.EventServer import EventServer
from icefarm.client.lib.BaseClient import BaseClient



from __future__ import annotations
from logging import Logger, LoggerAdapter
import threading
import atexit

import pyudev

from icefarm.utils.dev import *
from icefarm.worker.device import Device

import typing
if typing.TYPE_CHECKING:
    from icefarm.worker import Config, EventSender, WorkerDatabase

class ManagerLogger(LoggerAdapter):
    def __init__(self, logger, extra=None):
        super().__init__(logger, extra)

    def process(self, msg, kwargs):
        return f"[DeviceManager] {msg}", kwargs

class DeviceManager:
    """Tracks device events and routes them to their corresponding Device object. Also listens to kernel
    device events to identify usbip disconnects."""
    def __init__(self, event_sender: EventSender, database: WorkerDatabase, config: Config, logger: Logger):
        self.config: Config = config
        self.logger: Logger = ManagerLogger(logger)
        self.event_sender: EventSender = event_sender
        self.database = database

        atexit.register(self.onExit)

        self._devs: dict[str, Device] = {}
        self._dev_lock = threading.Lock()

        self.exiting: bool = False

        context = pyudev.Context()
        monitor = pyudev.Monitor.from_netlink(context)
        monitor.filter_by("tty")
        monitor.filter_by("block")

        observer = pyudev.MonitorObserver(monitor, self.handleDevEvent, name="manager-userevents")
        observer.start()

        self.scan()

    def scan(self):
        """Trigger add events for devices that are already connected."""
        self.logger.info("Scanning for devices")
        context = pyudev.Context().list_devices()

        for dev in context:
            self.handleDevEvent("add", dev)

        self.logger.info("Finished scan")

    def handleDevEvent(self, action: str, dev: pyudev.Device):
        """Ensures that a device is related to pico2ice and reroutes the event to handleAddDevice or
        handleRemoveDevice."""
        if self.exiting:
            return

        if dev.properties.get("ID_VENDOR_ID") not in ["2e8a", "1209"]:
            return

        dev = dict(dev)

        serial = get_serial(dev)

        if not serial:
            return

        with self._dev_lock:
            device = self._devs.get(serial)

            if not device:
                self.database.addDevice(serial)
                device = Device(serial, self, self.event_sender, self.database, self.logger)
                self._devs[serial] = device

        thread = threading.Thread(target=lambda : device.handleDeviceEvent(action, dev), name="dev-event-handler")
        thread.start()

    def handleRequest(self, serial: str, event: str, contents: dict):
        with self._dev_lock:
            dev = self._devs.get(serial)

        if not dev:
            self.logger.warning(f"request for {event} on {serial} but device not found")
            return

        return dev.handleRequest(event, contents)

    def reserve(self, serial: str, kind: str, args: dict):
        with self._dev_lock:
            device = self._devs.get(serial)

        if not device:
            self.logger.error(f"device {serial} reserved but does not exist")
            return False

        return device.handleReserve(kind, args)

    def unreserve(self, serial: str):
        with self._dev_lock:
            dev = self._devs.get(serial)

        if not dev:
            return False

        return dev.handleUnreserve()

    def reboot(self, serial: str):
        with self._dev_lock:
            dev = self._devs.get(serial)

        if not dev:
            return False

        return dev.reboot()

    def delete(self, serial: str):
        with self._dev_lock:
            dev = self._devs.get(serial)

        if not dev:
            return False

        dev.handleExit()
        with self._dev_lock:
            if serial in self._devs:
                del self._devs[serial]

        context = pyudev.Context().list_devices()

        for dev in context:
            if get_serial(dict(dev)) == serial:
                self.handleDevEvent("add", dev)

        return True

    def onExit(self):
        """Callback for cleanup on program exit"""
        with self._dev_lock:
            devs = list(self._devs.values())

        for dev in devs:
            dev.handleExit()

        self.database.onExit()



from __future__ import annotations
from configparser import ConfigParser
import os

from icefarm.utils import config_else_env
from icefarm.utils import get_ip

class Config:
    # TODO do logging here
    def __init__(self, path=None):
        if path:
            if not os.path.exists(path):
                raise Exception("Config file does not exist")

            parser = ConfigParser()
            parser.read(path)
        else:
            parser = None

        self.worker_name: str= config_else_env("USBIPICE_WORKER_NAME", "Connection", parser, error=False)
        if not self.worker_name:
            self.worker_name = os.environ.get("HOSTNAME")
            print(f"WARNING: using {self.worker_name}")

        if not self.worker_name:
            raise Exception("USBIPICE_WORKER_NAME not set, no HOSTNAME")

        self.server_port: str = config_else_env("USBIPICE_SERVER_PORT", "Connection", parser, default="8081")
        self.virtual_server_port: str = config_else_env("USBIPICE_VIRTUAL_PORT", "Connection", parser, default="8081")
        self.control_server_url: str = config_else_env("USBIPICE_CONTROL_SERVER", "Connection", parser, error=False)
        if not self.control_server_url:
            print("WARNING: not logging to control")
        self.virtual_ip: str = config_else_env("USBIPICE_VIRTUAL_IP", "Connection", parser, error=False)
        if not self.virtual_ip:
            self.virtual_ip = get_ip()
            print(f"WARNING: using {self.virtual_ip}")

        self.libpg_string= os.environ.get("USBIPICE_DATABASE")
        if not self.libpg_string:
            raise Exception("Environment variable USBIPICE_DATABASE not configured. Set this to a libpg \
            connection string to the database. If using sudo .venv/bin/worker, you may have to use the ENV= sudo arguments.")

        self.default_firmware_path = config_else_env("USBIPICE_DEFAULT", "Firmware", parser)
        self.pulse_firmware_path = config_else_env("USBIPICE_PULSE_COUNT", "Firmware", parser)



from __future__ import annotations
import logging
import threading
import json

import psycopg
from flask_socketio import SocketIO

from .Database import Database

class EventSenderLogger(logging.LoggerAdapter):
    def __init__(self, logger, extra=None):
        super().__init__(logger, extra)

    def process(self, msg, kwargs):
        return f"[EventSender] {msg}", kwargs

class SessionLogger(logging.LoggerAdapter):
    def __init__(self, logger, client_id, extra = None):
        super().__init__(logger, extra)
        self.client_id = client_id

    def process(self, msg, kwargs):
        return f"[{self.client_id}] {msg}", kwargs

class Session:
    def __init__(self, socketio: SocketIO, event_sender, logger: logging.Logger, client_id: str):
        self.socketio = socketio
        self.event_sender = event_sender
        self.logger = SessionLogger(logger, client_id)
        self.client_id = client_id

        self.sock_id = None
        self.message_queue = []

        self.lock = threading.Lock()
        self.timeout = None

        self.startTimeout()

    def startTimeout(self, time: int=60):
        def timeout():
            self.logger.error("client did not connect in time")
            self.event_sender.endSession(self.client_id)

        with self.lock:
            self.timeout = threading.Timer(time, timeout)
            self.timeout.daemon = True
            self.timeout.name = f"socket-session-{self.client_id}-timeout-monitor"
            self.timeout.start()

    def stopTimeout(self):
        with self.lock:
            if self.timeout:
                self.timeout.cancel()

    def send(self, data: str):
        with self.lock:
            self.message_queue.append(data)
        self.flush()

    def setSocket(self, sock_id):
        with self.lock:
            self.sock_id = sock_id
        self.logger.info("socket connected")

        self.stopTimeout()
        self.flush()

    def removeSocket(self):
        with self.lock:
            self.sock_id = None
        self.logger.info("socket disconnected")

        self.startTimeout()

    def flush(self):
        with self.lock:
            if not self.message_queue:
                return

            if not self.sock_id:
                self.logger.warning("no socket to flush to")
                return

            messages, self.message_queue = self.message_queue, []
            sock_id = self.sock_id

        for message in messages:
            try:
                self.socketio.emit("event", message, to=sock_id)
                self.socketio.sleep(0)
            except Exception:
                self.logger.warning("socket disconnected during flush")
                with self.lock:
                    self.message_queue.append(message)
                    return

        self.logger.debug(f"flushed {len(messages)} events")

class EventSender(Database):
    def __init__(self, socketio: SocketIO, dburl: str, logger: logging.Logger):
        super().__init__(dburl)
        self.socketio = socketio
        self.logger = EventSenderLogger(logger)

        self.sessions: dict[str, Session] = {}
        self.lock = threading.Lock()

    def startSession(self, client_id):
        with self.lock:
            if client_id not in self.sessions:
                self.sessions[client_id] = Session(self.socketio, self, self.logger, client_id)

            return self.sessions.get(client_id)

        self.logger.info(f"started session {client_id}")

    def addSocket(self, sock_id, client_id: str):
        session = self.startSession(client_id)
        session.setSocket(sock_id)

    def removeSocket(self, client_id):
        with self.lock:
            session = self.sessions.get(client_id)

        if session:
            session.removeSocket()
        else:
            self.logger.error(f"tried to socket for {client_id} but session does not exist")

    def endSession(self, client_id):
        with self.lock:
            self.sessions.pop(client_id, None)

    # TODO should probably use a database interface instead
    def __getReservationClientId(self, serial: str):
        """Returns the event server url for a device, None if there is none, or False on error."""
        try:
            with psycopg.connect(self.url) as conn:
                with conn.cursor() as cur:
                    cur.execute("SELECT * FROM get_device_callback(%s::varchar(255))", (serial,))
                    data = cur.fetchall()
        except Exception:
            self.logger.warning(f"failed to get device callback for serial {serial}")
            return False

        if not data:
            # no reservation
            return None

        return data[0][0]

    def sendClient(self, client_id: str, contents: str):
        session = self.startSession(client_id)
        session.send(contents)

    def sendSerial(self, serial, contents: str):
        client_id = self.__getReservationClientId(serial)
        if not client_id:
            self.logger.warning(f"tried to send event to {serial} but no reservation")
            return

        return self.sendClient(client_id, contents)

    def sendAll(self, contents: dict):
        if not (contents := self.__packageContents("meta", contents)):
            return False

        sessions = list(self.sessions.values())
        for session in sessions:
            session.send(contents)

    def __packageContents(self, serial: str, contents: dict):
        contents["serial"] = serial
        contents = {
            "serial": serial,
            "contents": contents
        }
        try:
            return json.dumps(contents)
        except Exception:
            return False

    def sendClientJson(self, serial: str, client_id: str, contents: dict) -> bool:
        contents = self.__packageContents(serial, contents)
        if not contents:
            return False

        self.sendClient(client_id, contents)
        return True

    def sendSerialJson(self, serial: str, contents: dict) -> bool:
        contents = self.__packageContents(serial, contents)
        if not contents:
            return False

        self.sendSerial(serial, contents)
        return True



import uuid
import threading
import math
from collections import Counter
from collections.abc import Generator
from abc import ABC
from typing import Any

from icefarm.utils import MappedQueues

class Evaluation:
    """
    Circuit evaluation to be sent to the worker. Note that providing
    multiple serials to one Evaluation will result in faster evaluations than
    providing multiple Evaluations of the same bitstream.
    """
    def __init__(self, serials: set[str]):
        if not serials:
            raise Exception("Serials cannot be empty")

        self.serials = frozenset(serials)
        self.id = str(uuid.uuid4())

    def __eq__(self, other):
        return self.id  == other.id

    def __hash__(self):
        return hash(self.id)

class EvaluationBundle:
    """
    Bundles Evaluations into efficient batches that can be sent gradually rather than
    all at once.
    """
    def __init__(self, evaluations: list[Evaluation], batch_size):
        self.queue = MappedQueues()
        for evaluation in evaluations:
            self.queue.append(evaluation.serials, evaluation)

        self.evaluation_lookup = {evaluation.id : evaluation for evaluation in evaluations}

        self.batch_size = batch_size

        self.empty = False
        self.id = str(uuid.uuid4())

    def __next__(self) -> dict[set[str], list[Evaluation]]:
        """
        Produces the next batch. Batches are produced so that each Evaluation with multiple serials are sent in the same batch
        and the maximum evaluations are sent without exceeding the batch size amount of circuits for any serial.
        """
        serial_amounts = Counter()
        batch = {}

        search_order = sorted(self.queue.keys(), key=len)
        for serials in search_order:
            slots = min(self.batch_size - serial_amounts[serial] for serial in serials)
            if not slots:
                continue

            commands = self.queue.pop(serials, slots)
            batch[serials] = commands
            for serial in serials:
                serial_amounts[serial] += len(commands)

        if not batch:
            self.empty = True
            raise StopIteration

        return batch

    def __iter__(self):
        return self

class AbstractBatchFactory(ABC):
    """
    Produces batches for client consumption.
    """
    def __init__(self, bundle: EvaluationBundle):
        super().__init__()
        self.bundle = bundle
        self.results: list[tuple[Evaluation, dict]] = []
        self.result_cv = threading.Condition()

        self.awaiting_results: dict[str, set] = {}

    def processResult(self, serial: str, evaluation_id: str, result: Any):
        with self.result_cv:
            if serial not in self.awaiting_results:
                return

            if evaluation_id not in self.awaiting_results[serial]:
                return

            self.awaiting_results[serial] -= {evaluation_id}
            self.results.append((serial, evaluation_id, result))
            self.result_cv.notify_all()

    def getResults(self) -> Generator[tuple[str, Evaluation, Any]]:
        """Produces results as they received by processResult."""
        while True:
            with self.result_cv:
                if not self.results:
                    self.result_cv.wait_for(lambda : self.results)

                for serial, evaluation_id, result in self.results:
                    evaluation = self.bundle.evaluation_lookup.get(evaluation_id)
                    yield serial, evaluation, result

                self.results = []

                if self.bundle.empty and not any(self.awaiting_results.values()):
                    return

    def _addBatch(self, batch: dict[set[str], list[Evaluation]]):
        """Adds batch to awaiting results."""
        for serials, evaluations in batch.items():
            for serial in serials:
                if serial not in self.awaiting_results:
                    self.awaiting_results[serial] = set()

                self.awaiting_results[serial] |= set(evaluation.id for evaluation in evaluations)

    def getBatches(self) -> Generator[dict[set[str], list[Evaluation]]]:
        """Produces batches for client to consume."""
        for batch in self.bundle:
            with self.result_cv:
                if not self._readyForBatch():
                    self.result_cv.wait_for(self._readyForBatch)

                self._addBatch(batch)
                yield batch

    def _readyForBatch(self) -> bool:
        """Whether the next batch should be made available for consumption."""

class QuickBatchFactory(AbstractBatchFactory):
    """
    Sends all batches to workers without any delay.
    """
    def _readyForBatch(self):
        return True

class PatientBatchFactory(AbstractBatchFactory):
    """
    Waits until all workers have finished processing a batch
    before sending a new one.
    """
    def _readyForBatch(self):
        return not any(self.awaiting_results.values())

class BalancedBatchFactory(AbstractBatchFactory):
    """
    Starts by sending an initial amount of batches, then
    waits until a threshold of results have been received
    before sending more batches.
    """
    def __init__(self, bundle, target_batches=4):
        super().__init__(bundle)
        self.target_batches = target_batches

    def _readyForBatch(self):
        if not any(self.awaiting_results.values()):
            return True

        highest_queued = max(map(len, self.awaiting_results.values()))
        batches = math.ceil(highest_queued / self.bundle.batch_size)
        return bool(self.target_batches - batches)



from logging import LoggerAdapter

from icefarm.utils import EventSender

class ControlEventSenderLogger(LoggerAdapter):
    def __init__(self, logger, extra=None):
        super().__init__(logger, extra)

    def process(self, msg, kwargs):
        return f"[ControlEventSender] {msg}", kwargs

class ControlEventSender(EventSender):
    def __init__(self, socketio, dburl, logger):
        super().__init__(socketio, dburl, ControlEventSenderLogger(logger))

    def sendDeviceReservationEnd(self, serial: str, client_id: str) -> bool:
        """Sends a reservation end event for serial."""
        if not self.sendClientJson(serial, client_id, {
            "event": "reservation end",
        }):
            self.logger.warning(f"failed to send reservation end to {client_id} for device {serial}")
        else:
            self.logger.info(f"sent reservation end to {client_id} for device {serial}")

    def sendDeviceFailure(self, serial: str, client_id: str) -> bool:
        """Sends a failure event for serial."""
        if not self.sendClientJson(serial, client_id, {
            "event": "failure",
        }):
            self.logger.warning(f"failed to send device failure to {client_id} for device {serial}")
        else:
            self.logger.info(f"sent device failure to {client_id} for device {serial}")

    def sendDeviceReservationEndingSoon(self, serial: str) -> bool:
        """Sends a reservation ending soon event for serial."""
        if not self.sendSerialJson(serial, {
            "event": "reservation ending soon",
        }):
            self.logger.warning(f"failed to send reservation ending soon to device {serial}")
        else:
            self.logger.info(f"sent reservation ending soon to device {serial}")



from __future__ import annotations
from collections.abc import Iterator
from dataclasses import dataclass
import json

import flask

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from icefarm.control import ControlDatabase

@dataclass
class WorkerRow:
    name: str
    url: str
    version: str
    shutting_down: str
    reservables_list: list[str]

    @property
    def reservables(self):
        if not self.reservables_list:
            return ""

        return ", ".join(self.reservables_list)

class Worker:
    def __init__(self, rows: Iterator[WorkerRow]):
        self.rows = list(rows)

    @property
    def online(self):
        return len(self.rows)

class DeviceRow:
    def __init__(self, serial, worker, status, client_id):
        self.serial = serial
        self.worker = worker
        self.status = status
        self.client_id = client_id

        js = "?json=" + json.dumps({
            "serials": [self.serial],
            "name": client_id
        })

        self.end_reservation = "./end" + js
        self.reboot = "./reboot" + js
        self.delete = "./delete" + js

class Device:
    def __init__(self, rows: Iterator[DeviceRow]):
        self.rows = list(rows)

    def filterStatus(self, status):
        return len(list(filter(lambda d : d.status == status, self.rows)))

    @property
    def total(self):
        return len(self.rows)

    @property
    def available(self):
        return self.filterStatus("available")

    @property
    def reserved(self):
        return self.filterStatus("reserved")

    @property
    def broken(self):
        return self.filterStatus("broken")

def build_page(database: ControlDatabase) -> str:
    try:
        if (worker_data := database.getWorkers()) is False:
            raise Exception

        worker = Worker(map(lambda r : WorkerRow(r["name"], f"http://{r["ip"]}:{r["port"]}", r["version"], r["shutting_down"], r["reservables"]), worker_data))

        if (device_data := database.getDevices()) is False:
            raise Exception

        device = Device(map(lambda r : DeviceRow(r["serial"], r["worker"], r["status"], r["client_id"]), device_data))

        return flask.render_template("home.html", workers=worker, devices=device)

    except Exception:
        return "Failed to fetch data"



"""
A collection of functions for interacting with device files.
"""
import os
import re
import subprocess

import pyudev

def get_serial(dev):
    """Obtains the serial from a dev file dict. Returns false if the dev file 
    is not related to pico2-ice."""
    devname = dev.get("DEVNAME")

    if not devname:
        return False

    if not re.match("/dev/", devname) or re.match("/dev/bus/", devname):
        return False

    id_model = dev.get("ID_MODEL")

    if id_model != "RP2350" and id_model != 'pico-ice' and id_model != 'Pico':
        return False

    serial = dev.get("ID_SERIAL_SHORT")

    if serial:
        return serial

    return False

def format_dev_file(udevinfo) -> str:
    """Formats a device file for printing."""
    id_serial = udevinfo.get("ID_SERIAL")
    usb_num = udevinfo.get("ID_USB_INTERFACE_NUM")
    dev_name = udevinfo.get("DEVNAME")
    dev_path = udevinfo.get("DEVPATH")
    return f"[{id_serial} : {usb_num} : {dev_name} : {dev_path}]"

def get_busid(dev_path: str) -> str:
    """Returns the bus from a devpath, or None. Obtains the bus from matching
    /usb1/.../() on DEVPATH."""
    # user format
    capture = re.search("/usb[0-9]/.*?/(.*?)([:/]|$)", dev_path)
    if capture:
        return capture.group(1)

    # kernel format
    capture = re.search("/usb[0-9]/([0-9]-(?:[0-9]|\\.)+)$", dev_path)
    if capture:
        return capture.group(1)
    return None

def mount(drive: str, loc: str, timeout: int=10) -> bool:
    """Mounts drive at location. Returns whether successful."""
    try:
        subprocess.run(["sudo", "mount", drive, loc], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, timeout=timeout, check=True)
    except Exception:
        return False

    return True

def umount(loc: str) -> bool:
    """Unmounts at location. Returns whether successful."""
    try:
        subprocess.run(["sudo", "umount", loc], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
    except Exception:
        return False

    return True

def send_bootloader(path: str, timeout: int=10):
    """Connects with picocom using a 1200 baud at path. Returns whether successful."""
    try:
        subprocess.run(["sudo", "picocom", "--baud", "1200", path], timeout=timeout, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
    except Exception:
        return False

    return True

def get_devs():
    """Returns a dict mapping device serials to list of dev info dicts. This operation 
    looks through all available dev files and is intended to be only used once after reserving devices.
    If you are dealing with frequent dev file changes, you should use a pyudev MonitorObserver instead."""
    out = {}

    context = pyudev.Context().list_devices()
    for dev in context:
        values = dict(dev)
        serial = get_serial(values)

        if not serial:
            continue

        devname = values.get("DEVNAME")

        if not devname:
            continue

        if serial not in out:
            out[serial] = []

        out[serial].append(dev)
    return out

def get_dev_paths():
    """Returns a dict mapping device serials to list of dev paths. This operation 
    looks through all available dev files and is intended to be only used once after reserving devices.
    If you are dealing with frequent dev file changes, you should use a pyudev MonitorObserver instead."""
    out = get_devs()
    for key in out:
        items = map(lambda x : x.get("DEVNAME"), out[key])
        filtered = filter(lambda x : x, items)
        out[key] = list(filtered)
    
    return out

class FirmwareUploadFail(Exception):
    def __init__(self, *args):
        super().__init__(*args)

def upload_firmware(partition_path: str, mount_location: str, firmware_bytes: bytes, mount_timeout=30):
    """Mounts the partition at location and uploads firmware_bytes to the drive. Requires sudo 
    to write to drive.."""
    mounted = mount(partition_path, mount_location, timeout=mount_timeout)

    if not mounted:
        return False

    if os.listdir(mount_location) != ["INDEX.HTM", "INFO_UF2.TXT"]:
        umount(mount_location)
        return False

    try:
        with open(os.path.join(mount_location, "firmware.uf2"), "wb") as f:
            f.write(firmware_bytes)
    except Exception:
        umount(partition_path)
        raise FirmwareUploadFail()

    umount(partition_path)

    return True

def upload_firmware_path(partition_path: str, mount_location: str, firmware_path: str, mount_timeout=10):
    """Mounts the partition at location and copies firmware_path to the drive."""
    mounted = mount(partition_path, mount_location, timeout=mount_timeout)

    if not mounted:
        return False

    if os.listdir(mount_location) != ["INDEX.HTM", "INFO_UF2.TXT"]:
        umount(mount_location)
        return False

    try:
        subprocess.run(["sudo", "cp", firmware_path, mount_location], timeout=15, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except Exception:
        umount(partition_path)
        raise FirmwareUploadFail()

    umount(partition_path)

    return True



import asyncio
import inspect
import json
from functools import wraps

from flask import Response, jsonify, request
from socketio import AsyncServer

from .utils import json_to_args, typecheck

def inject_and_return_json(func):
    """Injects request json values into arguments. Uses argument names as the json key. Typechecks arguments,
    only classes are supported. Returns a status=400 if a key is missing or the typecheck fails.
    Returns status=200 on True and status=500 on false. Otherwise, returns flask.jsonify of the result."""
    parameter_strings = [] # func args as string
    parameters = inspect.signature(func).parameters.values()

    for param in parameters:
        parameter_strings.append(param.name)

    @wraps(func)
    def handler_wrapper(*args):
        if request.content_type == "application/json":
            try:
                js = request.get_json()
            except Exception:
                return Response(status=400)
        else:
            try:
                js = json.loads(request.args.get("json"))
            except Exception:
                return Response(status=400)

        args = json_to_args(js, parameter_strings)

        if not typecheck(func, args):
            return Response(status=400)

        res = func(*args)
        if res is True or res is None:
            return Response(status=200)
        if res is False:
            return Response(status=500)

        return jsonify(res)

    return handler_wrapper

def flask_socketio_adapter_connect(func):
    """Adapter to allow flask_socketio.SocketIO eventhandlers to use the same interface as
    socketio.AsyncServer for @socketio.on("connect"). This is
    achieved by injecting an empty environment argument."""
    @wraps(func)
    def event_handler(*args):
        if len(args) == 1:
            return func(request.sid, {}, args[0])

        return func(*args)

    return event_handler

def flask_socketio_adapter_on(func):
    """Adapter to allow flask_socketio.SocketIO eventhandlers to use the same interface as
    socketio.AsyncServer for @socketio.on events. If only one argument is passed, request.sid is
    injected as the first argument. For the connection event, you must use
    flask_socketio_adapter_connection instead."""
    @wraps(func)
    def event_handler(*args):
        if len(args) == 2:
            return func(*args)
        elif len(args) == 1:
            return func(request.sid, *args)

    return event_handler

class SyncAsyncServer(AsyncServer):
    """Adapter to allow flask_socketio.SocketIO to have the same interface as socketio.AsyncServer while
    running as an ASGI app"""
    def emit(self, event, data=None, to=None, room=None, skip_sid=None, namespace=None, callback=None, ignore_queue=False):
        return asyncio.run(super().emit(event, data, to, room, skip_sid, namespace, callback, ignore_queue))

    def sleep(self, seconds=0):
        return asyncio.run(super().sleep(seconds))



from __future__ import annotations

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from icefarm.client.lib import EventServer, Event

REGISTERED_METHODS = {}

class JsonMethodCall:
    """Allows attribute to be called with json arguments."""
    def __init__(self, name: str, args: list[str]):
        self.name = name
        self.parms = args

    def __call__(self, obj, data):
        args = list(map(data.get, self.parms))

        if None in args:
            return False

        if not hasattr(obj, self.name):
            return False

        fn = getattr(obj, self.name)

        return fn(*args)

def register(key, *args):
    class Register:
        def __init__(self, fn):
            self.fn = fn

        def __set_name__(self, owner, name):
            if owner not in REGISTERED_METHODS:
                REGISTERED_METHODS[owner] = {}

            REGISTERED_METHODS[owner][key] = JsonMethodCall(name, args)

            setattr(owner, name, self.fn)

    return Register

class AbstractEventHandler:
    def __init__(self, event_server: EventServer):
        self.event_server = event_server

    def exit(self):
        """Called on EventServer shutdown."""

    def sendEvent(self, event: Event):
        self.event_server.sendEvent(event)

    def handleEvent(self, event: Event):
        """Dispatches event to registered methods."""
        search = [type(self)]
        attr = None


        while search:
            type_ = search.pop(0)
            methods = REGISTERED_METHODS.get(type_)

            if not methods:
                continue

            attr = methods.get(event.event)

            if attr:
                break

            for cls in reversed(type_.__bases__):
                search.insert(0, cls)

        if not attr:
            return False

        return attr(self, event.contents)



from __future__ import annotations
from logging import Logger, LoggerAdapter
import threading
import time

import requests
import schedule

from icefarm.control import ControlDatabase

import typing
if typing.TYPE_CHECKING:
    from icefarm.control import ControlEventSender

# TODO get values from config
class HeartbeatConfig:
    def __init__(self):
        self.heartbeat_poll_seconds: str = 15
        self.timeout_poll_seconds: str = 15
        self.timeout_duration_seconds: str = 60
        self.reservation_poll_seconds: str = 30
        self.reservation_expiring_poll_seconds: str = 300
        self.reservation_expiring_notify_at_seconds: str = 20 * 60

class HeartbeatLogger(LoggerAdapter):
    def process(self, msg, kwargs):
        return f"[Heartbeat] {msg}", kwargs

class Heartbeat:
    def __init__(self, event_sender: ControlEventSender, database_url: str, config: HeartbeatConfig, logger: Logger):
        self.event_sender = event_sender
        self.logger = HeartbeatLogger(logger)
        self.database = ControlDatabase(database_url)
        self.config = config
        self.thread = None

    def start(self):
        self.__startHeartBeatWorkers()
        self.__startWorkerTimeouts()
        self.__startReservationTimeouts()
        self.__startReservationEndingSoon()

        def run():
            while True:
                schedule.run_pending()
                time.sleep(1)

        self.thread = threading.Thread(target=run, daemon=True, name="heartbeat")
        self.thread.start()

    def __startHeartBeatWorkers(self):
        def do():
            def run():
                workers = self.database.getWorkers()

                if not workers:
                    return

                for row in workers:
                    name = row["name"]
                    ip = row["ip"]
                    port = row["port"]

                    url = f"http://{ip}:{port}/heartbeat"
                    try:
                        req = requests.get(url, timeout=10)

                        if req.status_code != 200:
                            raise Exception
                    except Exception:
                        self.logger.error(f"{name} failed heartbeat check")
                    else:
                        if not self.database.heartbeatWorker(name):
                            self.logger.error(f"failed to update heartbeat for {name}")
                        else:
                            self.logger.debug(f"heartbeat success for {name}")

            threading.Thread(target=run, name="heartbeat-worker", daemon=True).start()

        schedule.every(self.config.heartbeat_poll_seconds).seconds.do(do)

    def __startWorkerTimeouts(self):
        def do():
            def run(timeout_dur=self.config.timeout_duration_seconds):
                data = self.database.getWorkerTimeouts(timeout_dur)
                if not data:
                    return

                for row in data:
                    self.event_sender.sendDeviceFailure(row["serial"], row["client_id"])
                    self.logger.info(f"Worker {row['worker']} failed; sent device failure for client {row['client_id']} device {row['serial']}")

            threading.Thread(target=run, name="heartbeat-worker-timeouts", daemon=True).start()

        schedule.every(self.config.timeout_poll_seconds).seconds.do(do)

    def __startReservationTimeouts(self):
        def do():
            def run():
                if not (data := self.database.getReservationTimeouts()):
                    return

                for row in data:
                    self.logger.info(f"Reservation for device {row['serial']} by client {row['client_id']} ended")

            threading.Thread(target=run, name="heartbeat-reservation-timeouts", daemon=True).start()

        schedule.every(self.config.reservation_poll_seconds).seconds.do(do)

    def __startReservationEndingSoon(self):
        def do():
            def run(notify_at=self.config.reservation_expiring_notify_at_seconds):
                if not (data := self.database.getReservationEndingSoon(notify_at)):
                    return

                for serial in data:
                    self.event_sender.sendDeviceReservationEndingSoon(serial)

            threading.Thread(target=run, name="heartbeat-reservation-ending-soon", daemon=True).start()

        schedule.every(self.config.reservation_expiring_poll_seconds).seconds.do(do)



from __future__ import annotations
import threading
from logging import Logger, LoggerAdapter

from icefarm.utils import typecheck
from icefarm.utils.dev import *
from icefarm.worker.device import Device

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from icefarm.worker import WorkerDatabase, Config, DeviceEventSender

class EventMethod:
    def __init__(self, method, parms):
        self.method = method
        self.parms = parms

    def __call__(self, device, data):
        args = list(map(data.get, self.parms))

        if None in args:
            return

        if not typecheck(self.method, (device, *args)):
            return

        return self.method(device, *args)

class StateLogger(LoggerAdapter):
    def __init__(self, logger: Logger, state: str, extra = None):
        self.state = state
        super().__init__(logger, extra={})

    def process(self, msg, kwargs):
        return f"[{self.state}] {msg}", kwargs

class AbstractState:
    methods = {}

    def __init__(self, device: Device):
        """NOTE: switch CANNOT be called inside __init__(). This will result
        in the lock for Device being a acquired a second time. If this behavior
        is needed, use start() instead."""
        self.device: Device = device

        name = type(self).__name__
        self.logger: Logger = StateLogger(self.device.logger, name)

        self._switching: bool = False
        self.switching_lock: threading.RLock = threading.RLock()

        self.logger.info(f"state is now {name}")

    def start(self):
        """Called after the AbstractState is initialized, for
        actions that may result in a switch call. Note that handleExit
        may be called during this step."""

    def handleAdd(self, dev: dict):
        """Called on ADD device event."""

    def handleRemove(self, dev: dict):
        """Called on REMOVE device event."""

    def handleRequest(self, event, json):
        """Calls method event from the methods dictionary, using the arguments it was registered with
        as keys for the json."""
        methods = AbstractState.methods.get(type(self))

        if not methods:
            return

        method = methods.get(event)

        if method:
            return method(self, json)

        return

    def reboot(self):
        """Attempt to restore device state after request by client by switching to the"""
        self.logger.warning("Got reboot request but not implemented for this state")

    def handleExit(self):
        """Cleanup"""

    def switch(self, state_factory):
        """Switches the Device's state to a new one. This happens by first calling
        exit on the existing state. After the existing state has exited, the
        state factory is called and the result is set as the state. Subsequent calls
        to switch from the original state object are ignored."""
        with self.switching_lock:
            if self._switching:
                return

            self._switching = True

        return self.device.switch(state_factory)

    @classmethod
    def register(cls, event, *args):
        # TODO
        # update this for sockets
        # test typechecking with sockets
        """Adds a method to the methods dictionary, which allows it to be called
        using the handleEvent function with event=event. These arguments specify which json
        key should be used to get the value of that positional argument when handleEvent is called.
        The values passed in from the client are typechecked. Currently, only type and list[type]
        are supported. Files should be sent as cp437 encoded bytes. If the file is needed later, it
        should be saved under self.getDevice().getMediaPath(). Parameters without types are treated as Any.

        Ex.
        >>> class ExampleDevice:
                @AbstractState.register("add", "value 1", "value 2")
                def addNumbers(self, a: int, b: int):
                    self.getLogger().info(a + b)
        >>> client.requestWorker("serial", {
                "serial": "ABCDEF",
                "value 1": 1,
                "value 2": 2
        [ABCDEF] 3
        """
        class Reg:
            def __init__(self, fn):
                self.fn = fn

            def __set_name__(self, owner, name):
                if owner not in cls.methods:
                    cls.methods[owner] = {}

                if name in cls.methods[owner]:
                    raise Exception(f"{event} already registered")

                cls.methods[owner][event] = EventMethod(self.fn, args)
                setattr(owner, name, self.fn)

        return Reg

    @property
    def serial(self) -> str:
        return self.device.serial

    @property
    def database(self) -> WorkerDatabase:
        return self.device.database

    @property
    def device_event_sender(self) -> DeviceEventSender:
        return self.device.device_event_sender

    @property
    def config(self) -> Config:
        return self.device.config

    @property
    def switching(self):
        """Whether the Device is currently switching states"""
        with self.switching_lock:
            return self._switching



from __future__ import annotations
from logging import Logger
from typing import List
from itertools import groupby
import threading

from icefarm.client.lib import BaseAPI, EventServer, AbstractEventHandler, register
from icefarm.client.lib.utils import AvailabilityWaiter

class BaseClientEventHandler(AbstractEventHandler):
    """
    Updates the available serials on a client as reservations end and devices fail. Provides initialization
    detection for devices.
    """
    def __init__(self, event_server, client: BaseAPI):
        super().__init__(event_server)
        self.client = client
        self.recently_added_serials = []
        self.awaiting_serials = set()
        self.cond = threading.Condition()

    @register("reservation end", "serial")
    def handleReservationEnd(self, serial: str):
        self.client.removeSerial(serial)
        self.client.logger.warning(f"reservation for {serial} ended")

    @register("failure", "serial")
    def handleFailure(self, serial: str):
        self.client.removeSerial(serial)
        self.client.logger.warning(f"device {serial} failed")

    @register("initialized", "serial")
    def handleInitialization(self, serial):
        with self.cond:
            self.awaiting_serials.discard(serial)
            if not self.awaiting_serials:
                self.cond.notify_all()

    def waitUntilInitilized(self, serials):
        """Returns once 'initalized' events have been received for all the serials."""
        self.awaiting_serials = set(serials)

        with self.cond:
            for serial in self.recently_added_serials:
                self.awaiting_serials.pop(serial, None)

            if self.awaiting_serials:
                self.cond.wait_for(lambda : not self.awaiting_serials)

class BaseClient(BaseAPI):
    """Stitches the iCEFARM control server API and EventServer together to enable full management of devices."""
    def __init__(self, url: str, client_name: str, logger: Logger):
        super().__init__(url, client_name, logger)
        self.server = EventServer(client_name, [], logger)

        self.eh = BaseClientEventHandler(self.server, self)
        self.addEventHandler(self.eh)

        self.waiter = AvailabilityWaiter(self.server, self)
        self.addEventHandler(self.waiter)

        self.server.connectControl(url)

        # TODO track multiple initializations
        self.reservation_lock = threading.Lock()

    def addEventHandler(self, eh: AbstractEventHandler):
        self.server.addEventHandler(eh)

    def reserve(self, amount: int, kind: str, args: str, wait_for_available=False, available_timeout=None):
        """
        Reserves amount devices of type kind providing args to the worker when it is initilized. If wait_for_available,
        the client will wait until enough devices are available in the iCEFARM system. Otherwise, if there are not enough
        devices available, an error will be raised.
        """
        if self.available() < amount:
            if not wait_for_available:
                raise Exception("Not enough devices available")

            self.logger.warning("Not enough devices available, waiting for availability.")

            def raise_():
                raise Exception("Reservation timeout")

            if available_timeout:
                timer = threading.Timer(available_timeout, raise_)
                timer.daemon = True
                timer.name = "reserve-timeout-monitor"
                timer.start()

            self.waiter.waitForAmountAvailable(amount)
            if available_timeout:
                timer.cancel()

        with self.reservation_lock:
            serials = super().reserve(amount, kind, args)

            if not serials:
                return serials

            connected = []

            for serial in serials:
                info = self.getConnectionInfo(serial)

                if not info:
                    self.logger.error(f"could not get connection info for serial {serial}")

                self.server.connectWorker(info.url())
                connected.append(serial)

            self.eh.waitUntilInitilized(connected)
            return connected

    def reserveSpecific(self, serials, kind, args):
        """Reserves specific serials from the iCEFARM system. The serials must be available."""
        serials = super().reserveSpecific(serials, kind, args)

        if not serials:
            return serials

        connected = []

        for serial in serials:
            info = self.getConnectionInfo(serial)

            if not info:
                self.logger.error(f"could not get connection info for serial {serial}")

            self.server.connectWorker(info.url())
            connected.append(serial)

        self.eh.waitUntilInitilized(connected)
        return connected

    def removeSerial(self, serial):
        conn_info = self.getConnectionInfo(serial)
        super().removeSerial(serial)

        if not conn_info:
            return

        if not self.usingConnection(conn_info):
            self.server.disconnectWorker(conn_info.url())

    def requestWorker(self, serial: str, event: str, data: dict):
        """Sends data to socket of worker hosting serial. Note that the 'serial'
        field of the data will be override. If you want to duplicate requests across
        multiple serials, use requestBatchWorker instead.
        """
        info = self.getConnectionInfo(serial)
        if not info:
            return False

        return self.server.sendWorker(info.url(), "request", {
            "serial": serial,
            "event": event,
            "contents": data
        })

    def requestBatchWorker(self, serials: List[str], event: str, data: dict) -> List[str]:
        """Sends request to a list of serials. When the request is evaluated by the worker, the
        'serial' field is replaced. Returns serials included in requests to workers that failed.
        """
        failed_serials = []

        groups = groupby(serials, self.getConnectionInfo)

        for info, serials_iter in groups:
            batch_serials = list(serials_iter)

            if not info:
                self.logger.error(f"Could not get connection info for batch request serials: {batch_serials}")
                continue

            if not self.server.sendWorker(info.url(), "request", {
                "serial": batch_serials,
                "event": event,
                "contents": data
            }):
                self.logger.error(f"Failed to send request to worker {info.url()} for serials {batch_serials}")
                failed_serials.extend(batch_serials)

        return failed_serials

    def stop(self):
        self.server.exit()
        self.endAll()



from __future__ import annotations
from icefarm.utils import Database

class ControlDatabase(Database):

    def getDeviceWorkerUrl(self, serial: str) -> str:
        """Obtains the worker server url of the worker the device is located on."""
        if not (data := self.execute("SELECT * FROM get_device_worker(%s::varchar(255))", (serial,))):
            return False

        row = data[0]
        ip, port = row[0], row[1]
        return f"http://{ip}:{port}"

    def reserve(self, amount: int, clientname: str, reservation_type: str) -> dict:
        """Reserves amount devices for clientname. Returns as {serial, ip, serverport}"""
        return self.getData(
            "SELECT * FROM make_reservations(%s::int, %s::varchar(255), %s::varchar(255))", (amount, clientname, reservation_type),
            ["serial", "ip", "serverport"], stringify=["ip"]
        )

    def reserveSerials(self, client_id: str, serials: list[str], kind: str) -> dict:
        return self.getData(
            "SELECT * FROM make_specific_reservations(%s::varchar(255), %s::varchar(255)[], %s::varchar(255))", (client_id, serials, kind),
            ["serial", "ip", "serverport"], stringify=["ip"]
        )

    def extend(self, name: str, serials: list[str]) -> list[str]:
        """Extends the reservation time of the serials under the name of the client. Returns the extended serials"""
        if (data := self.execute("SELECT * FROM extend_reservations(%s::varchar(255), %s::varchar(255)[])", (name, serials))):
            return data[0]

        return False

    def extendAll(self, name: str) -> list[str]:
        """Extends the reservation time of all serials under the name of the client. Returns the extended serials."""
        if (data := self.execute("SELECT * FROM extend_all_reservations(%s::varchar(255))", (name,))):
            return data[0]

        return False

    def end(self, name: str, serials: list[str]):
        """Ends the reservation of serials under the name of the client.
        Returns as {serial, workerip, workerport}"""
        return self.getData(
            "select * from end_reservations(%s::varchar(255), %s::varchar(255)[])", (name, serials),
            ["serial", "workerip", "workerport"], stringify=["workerip", "workerport"]
        )

    def endAll(self, name: str):
        """Ends all of the reservations under the client name.
        Returns as {serial, workerip, workerport}"""
        return self.getData(
            "SELECT * FROM end_all_reservations(%s::varchar(255))", (name,),
            ["serial", "workerip", "workerport"], stringify=["workerip", "workerport"]
        )

    def getWorkers(self) -> dict:
        """Gets information about all of the workers, returns as a list of {name, ip, port}"""
        return self.getData(
            "SELECT * FROM worker", tuple(),
            ["name", "ip", "port", "heartbeat", "version", "reservables", "shutting_down"], stringify=["ip", "port"]
        )

    def getDevices(self) -> dict:
        """Returns current devices, as a list of {serial, worker, status}."""
        return self.getData(
            "SELECT * FROM device_reservations", tuple(),
            ["serial", "worker", "status", "client_id"], stringify=["status"]
        )

    def heartbeatWorker(self, name: str):
        """Updates the last heartbeat time on a worker to the current time"""
        return self.proc("CALL heartbeat_worker(%s::varchar(255))", (name,))

    def getWorkerTimeouts(self, timeout_dur: int) -> list:
        """Times out the workers that have not had a heartbeat in timeout_dur. Returns the
        timed out workers as a list of (serial, client_id, worker)."""
        return self.getData(
            "SELECT * FROM handle_worker_timeouts(%s::int)", (timeout_dur,),
            ["serial", "client_id", "worker"]
        )

    def getReservationEndingSoon(self, minutes: int) -> list[str]:
        """Gets reservations that are ending soon, returns the serials."""
        data = self.execute("SELECT * FROM get_reservations_ending_soon(%s::int)", (minutes,))
        if not data:
            return False

        return list(map(lambda x : x[0], data))

    def getReservationTimeouts(self) -> list[str]:
        """Gets reservations that have timed out, returns (serial, client_id)"""
        return self.getData(
            "SELECT * FROM handle_reservation_timeouts()", tuple(),
            ["serial", "client_id", "workerip", "workerport"], stringify=["workerip", "workerport"]
        )

    def getAmountAvailable(self) -> int:
        if not (data := self.execute("SELECT * FROM get_amount_available()", tuple())):
            return False

        return data[0][0]

    def getDevicesAvailable(self) -> list[str]:
        if (data := self.getData("SELECT * FROM get_available_devices()", tuple(), ["serial_ids"])) is False:
            return False

        return list(map(lambda x : x["serial_ids"], data))




from __future__ import annotations
from logging import Logger

from icefarm.utils import EventSender

class DeviceEventSender:
    """Allows for sending event notifications to client's event server, as well as sending
    instructions to worker's servers.."""
    def __init__(self, event_sender: EventSender, serial: str, logger: Logger):
        self.event_sender = event_sender
        self.serial = serial
        self.logger = logger

    def sendDeviceEvent(self, contents: dict) -> bool:
        if not self.event_sender.sendSerialJson(self.serial, contents):
            self.logger.error("failed to send event")
            return False

        return True

    def sendDeviceInitialized(self):
        return self.sendDeviceEvent({"event": "initialized"})

    def sendDeviceReservationEnd(self) -> bool:
        """Sends a reservation end event for serial."""
        return self.sendDeviceEvent({"event": "reservation end"})

    def sendDeviceFailure(self) -> bool:
        """Sends a failure event for serial."""
        return self.sendDeviceEvent({"event": "failure"})



"""Makes it possible for a specific device state to be requested from the client."""
from typing import List

state_value_checkers = {}
state_reservation_constructors = {}

def get_registered_reservables() -> List[str]:
    return list(state_reservation_constructors.keys())

def get_reservation_state_fac(state, kind, args):
    """Creates a factory for switching states based on a reservation request.
    Returns False if the event does not contain the correct arguments for the
    requested state or the state does not exist."""
    fn = state_reservation_constructors.get(kind)

    if not fn:
        return False

    return fn(state, args)

def reservable(name, *args: List[str]):
    """Makes an AbstractState available by reservation request under name. When reserve is called with
    this name, the device switches state to Cls(device: device.Device, *json_args), where json_args
    are obtained from using args as keys into the request dictionary. The class should be included in
    __init__.py to ensure the decorator is run.

    Ex.
    >>> @reservable("print device", "message")
        class ExampleState(device.state.core.AbstractState):
                def __init__(self, state, msg):
                    super().__init__()
                    self.logger.info(msg)
    >>> requests.get("{host}/reserve", json={
            "serial": "ASDF39TFDFG",
            "message": "hello!"
        })
    [ASDf39TFDFG] state is now ExampleState
    [ASDf39TFDFG] hello!
    """
    def res(cls):
        if name in state_reservation_constructors:
            raise Exception(f"reservable {name} is already registered (cls {cls})")

        def check_state(event):
            for arg in args:
                if not event.get(arg):
                    return False

            return True

        state_value_checkers[name] = check_state

        def make_state_fac(state, event):
            json_args = []

            for arg in args:
                json_args.append(event.get(arg))

            return lambda : cls(state, *json_args)

        state_reservation_constructors[name] = make_state_fac

        return cls

    return res



from __future__ import annotations
import threading
from typing import Dict
from collections.abc import Generator

from icefarm.client.lib.pulsecount import PulseCountBaseClient, PulseCountEventHandler, PulseCountEvaluation
from icefarm.client.lib import register
from icefarm.client.lib.utils import LoggerEventHandler, ReservationExtender
from icefarm.client.lib.BatchRequest import BalancedBatchFactory, EvaluationBundle

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from icefarm.client.lib.BatchRequest import AbstractBatchFactory, Evaluation

class PulseCountClient(PulseCountBaseClient):
    """Main client for pulse count experiments."""
    def __init__(self, url, client_name, logger, log_events=False):
        super().__init__(url, client_name, logger)

        self.addEventHandler(ReservationExtender(self.server, self, logger))
        if log_events:
            self.addEventHandler(LoggerEventHandler(self.server, logger))

        self.batch_factories: dict[str, AbstractBatchFactory] = {}

        class ResultHandler(PulseCountEventHandler):
            """Receives and processes results of circuit evaluations."""
            def __init__(self, event_server, client: PulseCountClient):
                super().__init__(event_server)
                self.client = client

            @register("results", "batch_id", "serial", "results")
            def results(self, batch_id: str, serial: str, results: Dict[str, int]):
                factory = self.client.batch_factories.get(batch_id)
                if not factory:
                    return

                for uid, pulses in results:
                    factory.processResult(serial, uid, pulses)

        self.addEventHandler(ResultHandler(self.server, self))

    def evaluateFactory(self, factory: AbstractBatchFactory) -> Generator[tuple[str, Evaluation, int]]:
        self.batch_factories[factory.bundle.id] = factory

        def evaluate_batches():
            for evaluations in factory.getBatches():
                for serial_group in evaluations.values():
                    self.evaluateBatch(factory.bundle.id, serial_group)

        threading.Thread(target=evaluate_batches, name="batch-sender").start()

        for result in factory.getResults():
            yield result

        del self.batch_factories[factory.bundle.id]

    def evaluateEvaluations(self, evaluations: list[PulseCountEvaluation], batch_size=5) -> Generator[tuple[str, Evaluation, int]]:
        return self.evaluateFactory(BalancedBatchFactory(EvaluationBundle(evaluations, batch_size)))

    def evaluateBitstreams(self, bitstreams: list[str], serials=None) -> Generator[tuple[str, str, int]]:
        """Sends bitstream filepaths to be evaluated by iCEFARM. If serials are not specified, bitstreams
        are evaluated on each reserved device. Results are received as (serial, filepath, pulses)."""
        if not serials:
            serials = self.getSerials()

        serials = set(serials)

        evaluations = [PulseCountEvaluation(serials, bitstream) for bitstream in bitstreams]
        for serial, evaluation, pulses in self.evaluateEvaluations(evaluations):
            yield serial, evaluation.filepath, pulses



from icefarm.worker.device.state.reservable.utils import get_reservation_state_fac, reservable, get_registered_reservables
from icefarm.worker.device.state.reservable.PulseCountState import PulseCountState



import os
import logging
import sys
import threading

from flask import Flask, request, send_file
from flask_socketio import SocketIO
from socketio import ASGIApp
from asgiref.wsgi import WsgiToAsgi

from icefarm.control import Control, Heartbeat, HeartbeatConfig, ControlEventSender
from icefarm.utils.web import SyncAsyncServer
from icefarm.utils.web import flask_socketio_adapter_connect, flask_socketio_adapter_on, inject_and_return_json

class ControlLogger(logging.LoggerAdapter):
    def __init__(self, logger, extra=None):
        super().__init__(logger, extra)

    def process(self, msg, kwargs):
        return f"[Control] {msg}", kwargs

def create_app(app: Flask, socketio: SocketIO | SyncAsyncServer, base_logger: logging.Logger):
    logger = ControlLogger(base_logger)

    DATABASE_URL = os.environ.get("USBIPICE_DATABASE")
    if not DATABASE_URL:
        raise Exception("USBIPICE_DATABASE not configured")

    sock_id_to_client_id = {}
    id_lock = threading.Lock()

    event_sender = ControlEventSender(socketio, DATABASE_URL, logger)
    control = Control(event_sender, DATABASE_URL, logger)

    heartbeat_config = HeartbeatConfig()
    heartbeat = Heartbeat(event_sender, DATABASE_URL, heartbeat_config, logger)
    heartbeat.start()

    @app.get("/")
    def get_app():
        return control.getApp()

    @app.get("/style.css")
    def style():
        return send_file("./assets/style.css")

    @app.get("/reboot")
    @inject_and_return_json
    def reboot(serials: list[str]):
        return control.reboot(serials)

    @app.get("/delete")
    @inject_and_return_json
    def delete(serials: list[str]):
        return control.delete(serials)

    @app.get("/available")
    def available():
        return control.getAmountAvailable()

    @app.get("/devices")
    def devices():
        return control.getDevicesAvailable()

    @app.get("/reserve")
    @inject_and_return_json
    def make_reservations(amount: int, name: str, kind: str, args: dict):
        return control.reserve(name, amount, kind, args)

    @app.get("/reserveserials")
    @inject_and_return_json
    def make_specific_reservations(name: str, kind: str, args: dict, serials: list[str]):
        return control.reserveSerials(name, serials, kind, args)

    @app.get("/extend")
    @inject_and_return_json
    def extend(name: str, serials: list):
        return control.extend(name, serials)

    @app.get("/extendall")
    @inject_and_return_json
    def extendall(name: str):
        return control.extendAll(name)

    @app.get("/end")
    @inject_and_return_json
    def end(name: str, serials: list):
        return control.end(name, serials)

    @app.get("/endall")
    @inject_and_return_json
    def endall(name: str):
        return control.endAll(name)

    @app.get("/log")
    @inject_and_return_json
    def log(name: str, logs: list):
        for row in logs:
            if len(row) != 2:
                continue

            level, msg = row[0], row[1]
            base_logger.log(level, f"[{name}@{request.remote_addr[0]}] {msg}")

        return True

    @socketio.on("connect")
    @flask_socketio_adapter_connect
    def connection(sid, environ, auth):
        client_id = auth.get("client_id")
        if not client_id:
            logger.warning("socket connection without client id")
            return

        logger.info(f"client {client_id} connected")

        with id_lock:
            sock_id_to_client_id[sid] = client_id

        event_sender.addSocket(sid, client_id)

    @socketio.on("disconnect")
    @flask_socketio_adapter_on
    def disconnect(sid, reason):
        with id_lock:
            client_id = sock_id_to_client_id.pop(sid, None)

        if not client_id:
            logger.warning("disconnected socket had no known client id")
            return

        logger.info(f"client {client_id} disconnected")

        event_sender.removeSocket(client_id)

def run_debug():
    SERVER_PORT = int(os.environ.get("USBIPICE_CONTROL_PORT", "8080"))

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.warning("Running in debug mode")

    app = Flask(__name__)

    socketio = SocketIO(app)
    create_app(app, socketio, logger)
    socketio.run(app, port=SERVER_PORT, allow_unsafe_werkzeug=True, host="0.0.0.0")

def run_uvicorn():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))

    app = Flask(__name__)
    socketio = SyncAsyncServer(async_mode="asgi")
    create_app(app, socketio, logger)
    app = WsgiToAsgi(app)

    return ASGIApp(socketio, app)


if __name__ == "__main__":
    run_debug()



import threading

from icefarm.worker.device.state.core import AbstractState, BrokenState, ReadyState
from icefarm.utils import check_default

class TestState(AbstractState):
    def __init__(self, state):
        super().__init__(state)
        self.lock = threading.Lock()
        self.exiting = False

        self.database.updateDeviceStatus(self.serial, "testing")

        self.timer = threading.Timer(30, lambda : self.switch(lambda : BrokenState(self.device)))
        self.timer.start()

    def handleAdd(self, dev):
        path = dev.get("DEVNAME")

        if not path:
            self.logger.warning("add event with no devname")
            return

        with self.lock:
            if self.exiting:
                return

            self.exiting = True

            if check_default(path):
                self.timer.cancel()
                self.switch(lambda : ReadyState(self.device))



from __future__ import annotations

from icefarm.client.lib import AbstractEventHandler, register, BaseClient
from icefarm.client.lib.BatchRequest import Evaluation

class PulseCountEvaluation(Evaluation):
    def __init__(self, serials, filepath):
        super().__init__(serials)
        self.filepath = filepath

class PulseCountEventHandler(AbstractEventHandler):
    @register("results", "batch_id", "serial", "results")
    # batch_id -> evaluation_id -> pulses
    def results(self, serial: str, results: dict[str, dict[str, list]]):
        """Called when ALL bitstreams have been evaluated. Results maps
        from the file parameter used in the request body to the
        pulse amount."""

class PulseCountBaseClient(BaseClient):
    """Provides access to pulse count specific control API methods."""
    def reserve(self, amount, wait_for_available=False, available_timeout=60):
        return super().reserve(amount, "pulsecount", {}, wait_for_available=wait_for_available, available_timeout=available_timeout)

    def reserveSpecific(self, serials: list[str]):
        return super().reserveSpecific(serials, "pulsecount", {})

    def evaluateBatch(self, batch_id: str, evaluations: list[PulseCountEvaluation]):
        """Sends a batch of PulseCountEvaluations to iCEFARM workers. The Evaluations must share
        the same set of serials."""
        if len(set([evaluation.serials for evaluation in evaluations])) != 1:
            raise Exception("Pulsecount evaluation commands contain different serials")

        files = {}
        for evaluation in evaluations:
            with open(evaluation.filepath, "rb") as f:
                files[evaluation.id] = f.read().decode("cp437")

        # files -> [data, id]
        return self.requestBatchWorker(list(evaluations[0].serials), "evaluate", {
            "files": files,
            "batch_id": batch_id
        })



import threading

from icefarm.worker.device.state.core import AbstractState, BrokenState

from icefarm.utils.dev import send_bootloader, upload_firmware_path, get_devs

class FlashState(AbstractState):
    def __init__(self, state, firmware_path, next_state_factory, timeout=None):
        super().__init__(state)
        self.firmware_path = firmware_path
        self.next_state_factory = next_state_factory
        self.timer = None

        if timeout:
            def do_timeout():
                self.logger.error("flashing timed out")
                self.switch(lambda : BrokenState(self.device))

            self.timer = threading.Timer(timeout, do_timeout)
            self.timer.daemon = True
            self.timer.name = f"{self.serial}-flash-timeout"
            self.timer.start()

    def start(self):
        devs = get_devs().get(self.serial)
        if not devs:
            return

        for file in devs:
            if self.switching:
                return

            self.handleAdd(file)

    def handleAdd(self, dev):
        devname = dev.get("DEVNAME")

        if not devname:
            self.logger.warning("add event with no devname")
            return

        if dev.get("SUBSYSTEM") == "tty":
            self.logger.debug(f"sending bootloader signal to {devname}")
            send_bootloader(devname)
            return

        if dev.get("DEVTYPE") == "partition":
            self.logger.debug(f"found bootloader candidate {devname}")

            uploaded = upload_firmware_path(devname, self.device.mount_path, self.firmware_path)

            if not uploaded:
                self.logger.error(f"failed to upload firmware to {devname}")
                if self.timer:
                    self.timer.cancel()
                self.switch(lambda : BrokenState(self.device))
                return

            if self.timer:
                self.timer.cancel()
            self.switch(self.next_state_factory)



"""Starts a graceful shutdown. Waits until the shutdown is complete to exit. Takes worker port as arg."""
import time
import sys
import threading
import socketio

port = sys.argv[1]

cv = threading.Condition()
shutdown = False

sio = socketio.Client()
@sio.event
def shutdown_complete(_):
    global shutdown
    with cv:
        shutdown = True
        cv.notify_all()

sio.connect(f"http://localhost:{port}", auth={"client_id":f"local_graceful_shutdown_{time.time()}"}, wait_timeout=10)
sio.emit("graceful_shutdown", "")

with cv:
    if not shutdown:
        cv.wait_for(lambda : shutdown)

sio.disconnect()